import os
import cv2
import time
import math
import torch
import torch.nn as nn
import requests
import warnings
from collections import deque
warnings.filterwarnings("ignore", category=UserWarning)

from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry
import mercantile
import numpy as np
import pandas as pd
import networkx as nx
import geopandas as gpd
from shapely import wkt
from tqdm import tqdm
import threading
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
import matplotlib.patches as mpatches
from matplotlib import cm
from shapely.strtree import STRtree
from scipy.stats import entropy

from concurrent.futures import ThreadPoolExecutor, as_completed, ProcessPoolExecutor
from shapely.geometry import Point, Polygon, box
from mapbox_vector_tile import decode
from vt2geojson.tools import vt_bytes_to_geojson
from vt2geojson.features import Layer, Feature

from ipyleaflet import Marker, Popup
from ipywidgets import HTML
from shapely.geometry import LineString, Point
from torchvision import transforms as T
from torch.utils.data import Dataset, DataLoader


from PIL import Image
from transformers import AutoImageProcessor, Mask2FormerForUniversalSegmentation
NUM_CLASSES = 65

CLS_DICT = {'0': 'Bird',
 '1': 'Ground-Animal',
 '2': 'Curb',
 '3': 'Fence',
 '4': 'Guard-Rail',
 '5': 'Barrier',
 '6': 'Wall',
 '7': 'Bike-Lane',
 '8': 'Crosswalk---Plain',
 '9': 'Curb-Cut',
 '10': 'Parking',
 '11': 'Pedestrian-Area',
 '12': 'Rail-Track',
 '13': 'Road',
 '14': 'Service-Lane',
 '15': 'Sidewalk',
 '16': 'Bridge',
 '17': 'Building',
 '18': 'Tunnel',
 '19': 'Person',
 '20': 'Bicyclist',
 '21': 'Motorcyclist',
 '22': 'Other-Rider',
 '23': 'Lane-Marking---Crosswalk',
 '24': 'Lane-Marking---General',
 '25': 'Mountain',
 '26': 'Sand',
 '27': 'Sky',
 '28': 'Snow',
 '29': 'Terrain',
 '30': 'Vegetation',
 '31': 'Water',
 '32': 'Banner',
 '33': 'Bench',
 '34': 'Bike-Rack',
 '35': 'Billboard',
 '36': 'Catch-Basin',
 '37': 'CCTV-Camera',
 '38': 'Fire-Hydrant',
 '39': 'Junction-Box',
 '40': 'Mailbox',
 '41': 'Manhole',
 '42': 'Phone-Booth',
 '43': 'Pothole',
 '44': 'Street-Light',
 '45': 'Pole',
 '46': 'Traffic-Sign-Frame',
 '47': 'Utility-Pole',
 '48': 'Traffic-Light',
 '49': 'Traffic-Sign-(Back)',
 '50': 'Traffic-Sign-(Front)',
 '51': 'Trash-Can',
 '52': 'Bicycle',
 '53': 'Boat',
 '54': 'Bus',
 '55': 'Car',
 '56': 'Caravan',
 '57': 'Motorcycle',
 '58': 'On-Rails',
 '59': 'Other-Vehicle',
 '60': 'Trailer',
 '61': 'Truck',
 '62': 'Wheeled-Slow',
 '63': 'Car-Mount',
 '64': 'Ego-Vehicle',
 '-1': 'Others'}

# For Visualisation
CLS_DICT_INT = {int(k): v for k, v in CLS_DICT.items()}
LABEL_TO_ID = {v: k for k, v in CLS_DICT.items()}          # string id
LABEL_TO_ID_INT = {v: int(k) for k, v in CLS_DICT.items()} # int id

perception_model_path = '../models'

perception = ['safety', 'lively', 'wealthy',
              'beautiful', 'boring', 'depressing']

perception_model_dict = {
    'safety': 'safety.pth',
    'lively': 'lively.pth',
    'wealthy': 'wealthy.pth',
    'beautiful': 'beautiful.pth',
    'boring': 'boring.pth',
    'depressing': 'depressing.pth',
}

# Client secret
session = requests.Session()
retry = Retry(connect=3, backoff_factor=0.5)
adapter = HTTPAdapter(max_retries=retry)
session.mount('http://', adapter)
session.mount('https://', adapter)

transform_img = T.Compose([
    T.ToTensor(),
    T.Resize((384, 384)),
    T.Normalize(
        mean=[0.485, 0.456, 0.406],
        std=[0.229, 0.224, 0.225])
])


def predict(model, img, device, transform=None):

    if transform:
        img = transform(img)

    img = img.view(1, 3, 384, 384)
    # inference
    if device == 'cuda':
        pred = model(img.cuda())
    else:
        pred = model(img)

    softmax = nn.Softmax(dim=1)
    pred = softmax(pred)[0][1].item()
    pred = round(pred*10, 2)

    return pred

def batch_predict(model, img_batch):

    pred = model(img_batch)

    softmax = nn.Softmax(dim=1)
    pred = [round(i.item()*10,2) for i in softmax(pred)[:,1]]
    
    return pred

def load_perception_models(perception_model_path, device):
    perception_models = []
    for p in perception:
        model_path = perception_model_path + "/" + perception_model_dict[p]
        model = torch.load(model_path, map_location=torch.device(device), weights_only=False)
        model = model.to(device)
        model.eval()
        perception_models.append(model)
    return perception_models

def addInstance(output_max):
    if isinstance(output_max, list):
       output_max = output_max[0]

    set_dictionary = {}
    for i in CLS_DICT.values():
        set_dictionary[i] = 0

    list_unique, list_counts = torch.unique(output_max['segmentation'].int(), return_counts=True)

    total = torch.sum(list_counts).item() 

    if -1 in list_unique:
        set_dictionary['Others'] = round(list_counts[0].item() / total, 5)
        list_unique = list_unique[1:]
        list_counts = list_counts[1:]

    matching_dict = {}

    for i, k in zip(range(len(output_max['segments_info'])), output_max['segments_info']):
        matching_dict[i] = int(k['label_id'])

    for i, k in zip(list_unique, list_counts):
        set_dictionary[CLS_DICT[str(matching_dict[i.item()])]] += round(k.item() / total, 5)
    
    set_dictionary['Visual Complexity'] = entropy(np.array(list(set_dictionary.values())), base=2)

    return set_dictionary

def addInstanceCounts(output_max):
    if isinstance(output_max, list):
       output_max = output_max[0]

    instance_dictionary = {}
    
    instance_dictionary = {}
    for i in CLS_DICT.values():
                instance_dictionary[str(i)] = 0
    
    # for each segment, draw its legend
    for segment in output_max['segments_info']:
        segment_id = segment['id']
        segment_label_id = str(segment['label_id'])
        instance_dictionary[CLS_DICT[segment_label_id]] += 1

    return instance_dictionary


def merge_zones_by_group(gdf, column):
    """Function to merge spatial dataframe according to a higher level of administrative grouping. 
    For example, this function can merge neighbourhoods into districts. 

    Args:
        gdf (gpd.GeoDataFrame): A spatial dataframe object with administrative zoning at a higher granularity.
        column (str): Specify column name to spatially merge administrative subzones groups.

    Returns:
        gpd.GeoDataFrame: Returns a spatial dataframe object with merged subzones. 
    """    
    unique_zones = list(gdf[column].unique())
    
    # Create empty city dataframe
    city_subzone = pd.DataFrame({'GID':[], 'geometry':[]})
    
    # Define internal merging function
    def merge_subzones(gdf, column, zone):
        polygons = gdf[gdf[column] == zone]
        geom = polygons['geometry'].unary_union.wkt
        df = pd.DataFrame({'GID':[zone], 'geometry':geom})
        df['geometry'] = df['geometry'].apply(lambda x: wkt.loads(x))
        gdf = gpd.GeoDataFrame(data=df, crs = gdf.crs, geometry=df['geometry'])
        return gdf
    
    # Iteratively add merged subzones
    for zone in unique_zones:
        merged = merge_subzones(gdf, column, zone)
        city_subzone = pd.concat([city_subzone, merged], axis=0)
        
    # Convert to GeoDataFrame
    city_subzone = gpd.GeoDataFrame(city_subzone, crs='epsg:4326', geometry = city_subzone['geometry'])
    
    return city_subzone

def sequence_download_image_in_tiles(gdf, api_key = ''):

    # Get bounds of shapefile
    bbox = list(gdf.geometry.bounds.values[0])

    # Obtain mercantile tiles for mapillary query
    tiles = list(mercantile.tiles(bbox[0], bbox[1], bbox[2], bbox[3], 14))

    # Initialise empty lists
    lat_list = []
    long_list = []
    image_list = []
    time_list = []
    angle_list = []
    tile_list = []
    creator_list = []
    sequence_list = []

    # Loop through tiles and download image meta info
    for tile in tqdm(tiles):
        tile_url = 'https://tiles.mapillary.com/maps/vtp/mly1_computed_public/2/{}/{}/{}'.format(tile.z,tile.x,tile.y)
        response = requests.get(tile_url, params={'access_token': api_key})
        
        geojson = request_to_geo_json(response, tile.x, tile.y, tile.z)
        
        # push to output geojson object if yes
        if geojson: 
            for feature in geojson['features']:
                # Filter images within bbox since tiles may extend beyond bounding box
                if not feature['properties']['is_pano']:
                    long_list.append(feature['geometry']['coordinates'][0])
                    lat_list.append(feature['geometry']['coordinates'][1])
                    image_list.append(feature['properties']['id'])
                    time_list.append(feature['properties']['captured_at'])
                    angle_list.append(feature['properties']['compass_angle'])
                    creator_list.append(feature['properties']['creator_id'])
                    sequence_list.append(feature['properties']['sequence_id'])
                    tile_list.append(f'{tile.z}/{tile.x}/{tile.y}')
                
    df = pd.DataFrame({'tile_id': tile_list, 
                       'image_id': image_list, 
                       'time':time_list, 
                       'angle': angle_list,
                       'creator': creator_list,
                       'sequence': sequence_list})
    image_gdf = gpd.GeoDataFrame(df, geometry = gpd.points_from_xy(long_list, lat_list), crs = 'EPSG:4326')

    image_gdf = image_gdf[image_gdf.within(gdf.geometry.values[0])]

    return image_gdf

# Process a tile with a session
def process_tile(tile, session, api_key, max_retries=3, retry_delay=1):
    tile_url = f'https://tiles.mapillary.com/maps/vtp/mly1_computed_public/2/{tile.z}/{tile.x}/{tile.y}'
    attempt = 0
    while attempt < max_retries:
        try:
            response = session.get(tile_url, params={'access_token': api_key})
            response.raise_for_status()
            break  # Exit loop if request is successful
        except requests.RequestException as e:
            attempt += 1
            print(f'Connection error for tile {tile.z}/{tile.x}/{tile.y} (Attempt {attempt}/{max_retries}): {e}')
            if attempt < max_retries:
                time.sleep(retry_delay)  # Wait before retrying
            else:
                print(f'Failed to fetch tile {tile.z}/{tile.x}/{tile.y} after {max_retries} attempts.')
                return []  # Return an empty list if all attempts fail

    geojson = request_to_geo_json(response, tile.x, tile.y, tile.z)
    
    data = []
    if geojson:
        for feature in geojson['features']:
            if not feature['properties']['is_pano']:
                data.append({
                    'long': feature['geometry']['coordinates'][0],
                    'lat': feature['geometry']['coordinates'][1],
                    'image_id': feature['properties']['id'],
                    'time': feature['properties']['captured_at'],
                    'angle': feature['properties']['compass_angle'],
                    'creator_id': feature['properties']['creator_id'],
                    'sequence_id': feature['properties']['sequence_id'],
                    'tile_id': f'{tile.z}/{tile.x}/{tile.y}'
                })
    return data

def parallel_download_image_in_tiles(gdf, api_key = ''):
    bbox = list(gdf.geometry.bounds.values[0])
    total_tiles = list(mercantile.tiles(bbox[0], bbox[1], bbox[2], bbox[3], 14))

    geometries = [box(*mercantile.bounds(tile)) for tile in total_tiles]

    # Create a GeoDataFrame
    tile_gdf = gpd.GeoDataFrame(data = {'tile_id':list(range(len(geometries)))},geometry=geometries, crs="EPSG:4326")
    tile_intersection = tile_gdf.overlay(gdf)
    tiles = [total_tiles[idx] for idx in tile_intersection['tile_id'].values]

    # Initialize lists
    lat_list = []
    long_list = []
    image_list = []
    time_list = []
    angle_list = []
    tile_list = []
    creator_list = []
    sequence_list = []

    results = []
    results_lock = threading.Lock()

    # Create a session object
    session = requests.Session()
    
    # Create and start threads
    # with tqdm(total=len(tiles)) as pbar:
    with ThreadPoolExecutor() as executor:
        future_to_tile = {executor.submit(process_tile, tile, session, api_key): tile for tile in tiles}
        for future in as_completed(future_to_tile):
            try:
                data = future.result()
                with results_lock:
                    results.extend(data)
            except Exception as e:
                print(f'Error occurred while processing tile: {e}')
                # pbar.update(1)

    # Close the session
    session.close()

    # Collect results from all threads
    for entry in results:
        long_list.append(entry['long'])
        lat_list.append(entry['lat'])
        image_list.append(entry['image_id'])
        time_list.append(entry['time'])
        angle_list.append(entry['angle'])
        creator_list.append(entry['creator_id'])
        sequence_list.append(entry['sequence_id'])
        tile_list.append(entry['tile_id'])

    df = pd.DataFrame({
        'tile_id': tile_list, 
        'image_id': image_list, 
        'time': time_list, 
        'angle': angle_list,
        'creator': creator_list,
        'sequence': sequence_list
    })

    image_gdf = gpd.GeoDataFrame(df, geometry=gpd.points_from_xy(long_list, lat_list), crs='EPSG:4326')
    image_gdf = image_gdf[image_gdf.within(gdf.geometry.values[0])]

    return image_gdf


def process_image(image_id, session, api_key, max_retries=3, retry_delay=1):

    url = f'https://graph.mapillary.com/{image_id}?fields=thumb_256_url'

    attempt = 0
    while attempt < max_retries:
        try:
            response = session.get(url, params={'access_token': api_key})
            response.raise_for_status()
            image_data = response.json()
            image_url = image_data['thumb_256_url']
    
            response = session.get(image_url, stream=True)
            response.raise_for_status()
            response = response.raw
        
            break  # Exit loop if request is successful
        except requests.RequestException as e:
            attempt += 1
            print(f'Connection error for image: {image_id} (Attempt {attempt}/{max_retries}): {e}')
            if attempt < max_retries:
                time.sleep(retry_delay)  # Wait before retrying
            else:
                print(f'Failed to fetch image: {image_id} after {max_retries} attempts.')
                return []  # Return an empty list if all attempts fail

    image_array = np.asarray(bytearray(response.read()), dtype=np.uint8)
    image = cv2.imdecode(image_array, cv2.IMREAD_COLOR)   
    
    # Convert BGR to RGB
    image_rgb = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    
    return image_id, image_rgb

def segment_image(img_id, image_rgb, processor, segment_model, device):
    # Preprocessing
    inputs = processor(images=image_rgb, return_tensors="pt")
    with torch.no_grad():
        pixel_values = inputs['pixel_values'].to(device)
        pixel_mask = inputs['pixel_mask'].to(device)
        outputs = segment_model(pixel_values=pixel_values, pixel_mask=pixel_mask)
    # Post-process
    out = processor.post_process_instance_segmentation(
        outputs, target_sizes=[image_rgb.shape[:-1]], threshold=0.5
    )
    return img_id, addInstance(out), addInstanceCounts(out)

def parallel_segment_images(image_gdf, processor, segment_model, api_key, device):
    # Initialise placeholders
    image_instances_dict = {}
    image_indicators_dict = {}

    results = {}
    results_lock = threading.Lock()

    # Create a session object
    session = requests.Session()
    
    failures = []
    # Create and start threads
    with ThreadPoolExecutor(max_workers=None) as executor:
        future_to_id = {executor.submit(process_image, image_id, session, api_key): image_id for image_id in image_gdf['image_id']}
        with tqdm(total=len(future_to_id), desc="Downloading images", unit="img") as pbar:
            for future in as_completed(future_to_id):
                image_id = future_to_id[future]
                try:
                    img_id, image_rgb = future.result()
                    with results_lock:
                        results[img_id] = image_rgb
                except Exception as e:
                    failures.append((image_id, e))
                finally:
                    pbar.update(1)

    # Close the session
    session.close()
    print('Downloaded images...')
    print('Starting segmentation')
    # Load model, etc. as before
    segment_model = segment_model.to(device)

    # Prepare results containers
    image_instances_dict = {}
    image_indicators_dict = {}

    futures = []
    with ThreadPoolExecutor(max_workers=None) as executor:  # adjust workers as needed
        for img_id, image_rgb in results.items():
            futures.append(executor.submit(
                segment_image, img_id, image_rgb, processor, segment_model, device
            ))
        for future in tqdm(as_completed(futures), total=len(futures)):
            img_id, indicators, instances = future.result()
            image_indicators_dict[img_id] = indicators
            image_instances_dict[img_id] = instances

    # Convert to dataframes
    indicators_df = pd.DataFrame.from_dict(image_indicators_dict, orient='index')
    instances_df = pd.DataFrame.from_dict(image_instances_dict, orient='index')

    # Merge dataframes
    merged = image_gdf.merge(indicators_df, how='left', left_on='image_id', right_on = indicators_df.index)
    merged = merged.merge(instances_df, how='left', left_on='image_id', right_on = indicators_df.index, suffixes=('_pixels', '_counts'))

    return merged

def parallel_batch_images(api_key, batch_size_ids):

    results = {}
    results_lock = threading.Lock()

    # Create a session object
    session = requests.Session()

    
    # Create and start threads
    # with tqdm(total=len(tiles)) as pbar:
    with ThreadPoolExecutor() as executor:
        future_to_tile = {executor.submit(process_image, image_id, session, api_key): image_id for image_id in batch_size_ids}
        for future in as_completed(future_to_tile):
            try:
                image_id, image_rgb = future.result()
                with results_lock:
                    results[image_id] = image_rgb
            except Exception as e:
                print(f'Error occurred while processing tile: {e}')
                # pbar.update(1)

    # Close the session
    session.close()
    
    return results.items()

def parallel_segment_images_in_batches(image_gdf, 
                                       processor,
                                       model, 
                                       api_key, 
                                       device, 
                                       batch_size=1000,
                                       gpu_batch=4,
                                       add_perception=False,
                                       perception_models=None):

    # Load images
    image_set = list(image_gdf['image_id'].unique())

    model = model.to(device)
    
    # Load Perception Models
    if add_perception:
        if perception_models is None:
            raise Exception('Please specify perception models')
        
        columns = ['image_id'] + [str(p)+"_score" for p in perception]
        perception_df = pd.DataFrame()

    # Initialise placeholders
    image_instances_dict = {}
    image_indicators_dict = {}

    for i in range(0, len(image_set), batch_size):
        batch_image_ids = image_set[i:i + batch_size]
        results = parallel_batch_images(api_key, batch_image_ids)
        print(f'Segmenting images: {i+batch_size}/{len(image_set)}')

        results_img_id = [img_id for img_id, img_rgb in results]
        results_img_rgb = [img_rgb for img_id, img_rgb in results]

        for i in range(0, len(results), gpu_batch):
            img_id_list = results_img_id[i:i + gpu_batch]
            img_rgb_list = results_img_rgb[i:i + gpu_batch]
            processed = [processor(images=img_rgb, return_tensors="pt") for img_rgb in img_rgb_list]
            value_tensor_list = torch.concat([value['pixel_values'] for value in processed]).to(device)
            mask_tensor_list = torch.concat([value['pixel_mask'] for value in processed]).to(device)
            # Inference
            outputs = model(pixel_values=value_tensor_list, pixel_mask=mask_tensor_list)
            out = processor.post_process_instance_segmentation(outputs, target_sizes=[img.shape[:-1] for img in img_rgb_list], threshold=0.5)

            instance_list = map(addInstance, out)
            instance_counts_list = map(addInstanceCounts, out)

            for img_id, instance_dict in zip(img_id_list, instance_list):
                image_instances_dict[img_id] = instance_dict

            for img_id, instance_counts_dict in zip(img_id_list, instance_counts_list):
                image_indicators_dict[img_id] = instance_counts_dict

            if add_perception:
                perception_preds = [batch_predict(model, value_tensor_list) for model in perception_models]
                new_df = pd.DataFrame({'image_id': img_id_list,
                                    'safety_score': perception_preds[0],
                                    'lively_score': perception_preds[1],
                                    'wealthy_score': perception_preds[2],
                                    'beautiful_score': perception_preds[3],
                                    'boring_score': perception_preds[4],
                                    'depressing_score': perception_preds[5]})

                perception_df = pd.concat([perception_df, pd.DataFrame(new_df)], ignore_index=True)

    # Convert to dataframes
    indicators_df = pd.DataFrame.from_dict(image_indicators_dict, orient='index')
    instances_df = pd.DataFrame.from_dict(image_instances_dict, orient='index')

    # Merge dataframes
    merged = image_gdf.merge(indicators_df, how='left', left_on='image_id', right_on = indicators_df.index)
    merged = merged.merge(instances_df, how='left', left_on='image_id', right_on = indicators_df.index, suffixes=('_counts', '_pixels'))

    if add_perception:
        merged = merged.merge(perception_df, how='left', on='image_id')
    
    return merged
    
class MyFeature(Feature):
    def __init__(self, x, y, z, obj, extent=4096):
        super().__init__(x, y, z, obj, extent=4096)
    
    @property
    def tiles_coordinates(self):     
        return self.obj['geometry']['coordinates'] 
    
    @property
    def geometry_type(self):
        return 'Point'
    
    @property
    def properties(self):
        return self.obj['properties']   
    
    def toGeoJSON(self):
        size = self.extent * 2 ** self.z
        x0 = self.extent * self.x
        y0 = self.extent * self.y

        def project_one(p_x, p_y):
            y2 = 180 - (p_y + y0) * 360. / size
            long_res = (p_x + x0) * 360. / size - 180
            lat_res = 360. / math.pi * math.atan(math.exp(y2 * math.pi / 180)) - 90
            return [long_res, lat_res]

        def project(coords):
            if all(isinstance(x, int) or isinstance(x, float) for x in coords):
                assert len(coords) == 2
                return project_one(coords[0], coords[1])
            return [project(l) for l in coords]

        coords = project(self.tiles_coordinates)
        geometry_type = self.geometry_type

        result = {
            "type": "Feature",
            "geometry": {
                "type": geometry_type,
                "coordinates": coords
            },
            "properties": self.properties
        }
        return result

class MyLayer(Layer):
    def __init__(self, x, y, z, name, obj):
        super().__init__(x, y, z, name, obj)
    
    def toGeoJSON(self):
        return {
            "type": "FeatureCollection",
            "features": [MyFeature(x=self.x, y=self.y, z=self.z, obj=image, extent=4096).toGeoJSON()
                         for image in self.obj['features']]
        }
    
def request_to_geo_json(response: bytes, x: int, y: int, z: int, layer='image') -> dict:
    """
    Make a GeoJSON from vector tile query
    :param content: request information from Mapillary vector tile query
    :param x: tile x coordinate.
    :param y: tile y coordinate.
    :param z: tile z coordinate.
    :param layer: image or sequence layer. Default: 'image'.
    :return: GeoJSON dictionary.
    """
    
    try:
        data = decode(response.content, y_coord_down=True)
        geojson = MyLayer(x=x, y=y, z=z, name=layer, obj=data[layer]).toGeoJSON()
        return geojson
    except:
        # No images found in tile --skipping--
        pass
    
def filter_by_alignment(G, gdf, limit=10):
    
    # Add bearings to graph edges
    G = add_edge_bearings(G, precision=1)
    
    # Project graph to enable distance computation
    gdf2 = project_gdf(gdf) # image_gdf
    near_edge = nearest_edges(G, gdf2.geometry.x, gdf2.geometry.y)
    
    # Create list of origin and destination tuples
    nearest_edge_list = []
    for edge in near_edge:
        nearest_edge_list.append(edge[0:2])
    
    # Create dataframe consisting of nearest edge, bearing, and back_bearing
    edges = graph_to_gdf(G, edges=True).reset_index()
    edges['nearest_edge'] = list(zip(edges.u, edges.v))
    edges = edges[['nearest_edge', 'bearing', 'back_bearing']]
    
    # Assign nearest edge joining column to projected graph
    gdf2['nearest_edge'] = nearest_edge_list
    
    # Merge dataframes on nearest edge to combine image angle and bearing information
    merged_df = gdf2.merge(edges, how = 'left', on = 'nearest_edge')
    
    # If angle difference less than 20, return image, else discard
    merged_df['aligned'] = merged_df.apply(lambda row: check_align(row['bearing'], row['back_bearing'], row['angle'], limit), axis=1)
    merged_df = merged_df[merged_df['aligned'] == True]
    merged_df = merged_df.to_crs(4326)
    
    return merged_df

def explode_linestrings(gdf):
    """Explode linestring into single pairs of coordinate sequences"""
    exploded_lines = []
    
    # Iterate over each row in the GeoDataFrame
    for _, row in gdf.iterrows():
        line = row.geometry
        coords = list(line.coords)
        
        # Create segments from coordinates
        for i in range(len(coords) - 1):
            segment = LineString([coords[i], coords[i + 1]])
            exploded_lines.append({'geometry': segment, **row.drop('geometry')})
    
    # Create a new GeoDataFrame from exploded lines
    exploded_gdf = gpd.GeoDataFrame(exploded_lines, geometry='geometry', crs=gdf.crs)
    exploded_gdf['exploded_edge_id'] = exploded_gdf.index
    return exploded_gdf


def filter_by_alignment_finegrained(image_gdf_proj, tract_edges, limit=10):
    exploded_gdf = explode_linestrings(tract_edges)
    exploded_gdf['bearing'] = exploded_gdf['geometry'].apply(lambda x: calculate_bearing(x.coords[0][0], x.coords[0][1], x.coords[1][0], x.coords[1][1]))
    exploded_gdf['back_bearing'] = exploded_gdf['bearing'].apply(lambda x: (x - 180) if x >= 180 else (x + 180))

    exploded_gdf_proj = project_gdf(exploded_gdf)
    nearest_edge = gpd.sjoin_nearest(image_gdf_proj, exploded_gdf_proj, how='inner')
    nearest_edge = nearest_edge.drop_duplicates(subset='image_id', keep='first')

    # If angle difference less than limit, return image, else discard
    nearest_edge['aligned'] = nearest_edge.apply(lambda row: check_align(row['bearing'], row['back_bearing'], row['angle'], limit), axis=1)
    nearest_edge = nearest_edge[nearest_edge['aligned'] == True]
    nearest_edge.crs = image_gdf_proj.crs
    nearest_edge = nearest_edge.to_crs(4326)



# Attach images to network and assign image bearing
def calculate_bearing(lat1, lng1, lat2, lng2):
    """
    Calculate the compass bearing(s) between pairs of lat-lng points.

    Vectorized function to calculate (initial) bearings between two points'
    coordinates or between arrays of points' coordinates. Expects coordinates
    in decimal degrees. Bearing represents angle in degrees (clockwise)
    between north and the geodesic line from point 1 to point 2.

    Parameters
    ----------
    lat1 : float or numpy.array of float
        first point's latitude coordinate
    lng1 : float or numpy.array of float
        first point's longitude coordinate
    lat2 : float or numpy.array of float
        second point's latitude coordinate
    lng2 : float or numpy.array of float
        second point's longitude coordinate

    Returns
    -------
    bearing : float or numpy.array of float
        the bearing(s) in decimal degrees
    """
    # get the latitudes and the difference in longitudes, in radians
    lat1 = np.radians(lat1)
    lat2 = np.radians(lat2)
    d_lng = np.radians(lng2 - lng1)

    # calculate initial bearing from -180 degrees to +180 degrees
    y = np.sin(d_lng) * np.cos(lat2)
    x = np.cos(lat1) * np.sin(lat2) - np.sin(lat1) * np.cos(lat2) * np.cos(d_lng)
    initial_bearing = np.degrees(np.arctan2(y, x))

    # normalize to 0-360 degrees to get compass bearing
    return initial_bearing % 360


def add_edge_bearings(G, precision=1):
    """
    Add compass `bearing` attributes to all graph edges.

    Vectorized function to calculate (initial) bearing from origin node to
    destination node for each edge in a directed, unprojected graph then add
    these bearings as new edge attributes. Bearing represents angle in degrees
    (clockwise) between north and the geodesic line from the origin node to
    the destination node. Ignores self-loop edges as their bearings are
    undefined.

    Parameters
    ----------
    G : networkx.MultiDiGraph
        unprojected graph
    precision : int
        decimal precision to round bearing

    Returns
    -------
    G : networkx.MultiDiGraph
        graph with edge bearing attributes
    """

    # extract edge IDs and corresponding coordinates from their nodes
    uvk = [(u, v, k) for u, v, k in G.edges if u != v]
    x = G.nodes(data="x")
    y = G.nodes(data="y")
    coords = np.array([(y[u], x[u], y[v], x[v]) for u, v, k in uvk])

    # calculate bearings then set as edge attributes
    bearings = calculate_bearing(coords[:, 0], coords[:, 1], coords[:, 2], coords[:, 3])
    back_bearing = calculate_bearing(coords[:, 2], coords[:, 3], coords[:, 0], coords[:, 1])
    val = zip(uvk, bearings.round(precision))
    back_val = zip(uvk, back_bearing.round(precision))
    nx.set_edge_attributes(G, dict(val), name="bearing")
    nx.set_edge_attributes(G, dict(back_val), name="back_bearing")

    return G

def nearest_edges(G, X, Y, interpolate=None, return_dist=False):
    """
    Find the nearest edge to a point or to each of several points.

    If `X` and `Y` are single coordinate values, this will return the nearest
    edge to that point. If `X` and `Y` are lists of coordinate values, this
    will return the nearest edge to each point.

    If `interpolate` is None, search for the nearest edge to each point, one
    at a time, using an r-tree and minimizing the euclidean distances from the
    point to the possible matches. For accuracy, use a projected graph and
    points. This method is precise and also fastest if searching for few
    points relative to the graph's size.

    For a faster method if searching for many points relative to the graph's
    size, use the `interpolate` argument to interpolate points along the edges
    and index them. If the graph is projected, this uses a k-d tree for
    euclidean nearest neighbor search, which requires that scipy is installed
    as an optional dependency. If graph is unprojected, this uses a ball tree
    for haversine nearest neighbor search, which requires that scikit-learn is
    installed as an optional dependency.

    Parameters
    ----------
    G : networkx.MultiDiGraph
        graph in which to find nearest edges
    X : float or list
        points' x (longitude) coordinates, in same CRS/units as graph and
        containing no nulls
    Y : float or list
        points' y (latitude) coordinates, in same CRS/units as graph and
        containing no nulls
    interpolate : float
        spacing distance between interpolated points, in same units as graph.
        smaller values generate more points.
    return_dist : bool
        optionally also return distance between points and nearest edges

    Returns
    -------
    ne or (ne, dist) : tuple or list
        nearest edges as (u, v, key) or optionally a tuple where `dist`
        contains distances between the points and their nearest edges
    """
    is_scalar = False
    if not (hasattr(X, "__iter__") and hasattr(Y, "__iter__")):
        # make coordinates arrays if user passed non-iterable values
        is_scalar = True
        X = np.array([X])
        Y = np.array([Y])

    if np.isnan(X).any() or np.isnan(Y).any():  # pragma: no cover
        raise ValueError("`X` and `Y` cannot contain nulls")
    geoms = project_gdf(graph_to_gdf(G, edges=True))["geometry"]

    # if no interpolation distance was provided
    if interpolate is None:

        # build the r-tree spatial index by position for subsequent iloc
        rtree = RTreeIndex()
        for pos, bounds in enumerate(geoms.bounds.values):
            rtree.insert(pos, bounds)

        # use r-tree to find possible nearest neighbors, one point at a time,
        # then minimize euclidean distance from point to the possible matches
        ne_dist = list()
        for xy in zip(X, Y):
            dists = geoms.iloc[list(rtree.nearest(xy, num_results=10))].distance(Point(xy))
            ne_dist.append((dists.idxmin(), dists.min()))
        ne, dist = zip(*ne_dist)

    # otherwise, if interpolation distance was provided
    else:

        # interpolate points along edges to index with k-d tree or ball tree
        uvk_xy = list()
        for uvk, geom in zip(geoms.index, geoms.values):
            uvk_xy.extend((uvk, xy) for xy in utils_geo.interpolate_points(geom, interpolate))
        labels, xy = zip(*uvk_xy)
        vertices = pd.DataFrame(xy, index=labels, columns=["x", "y"])

        if projection.is_projected(G.graph["crs"]):
            # if projected, use k-d tree for euclidean nearest-neighbor search
            if cKDTree is None:  # pragma: no cover
                raise ImportError("scipy must be installed to search a projected graph")
            dist, pos = cKDTree(vertices).query(np.array([X, Y]).T, k=1)
            ne = vertices.index[pos]

        else:
            # if unprojected, use ball tree for haversine nearest-neighbor search
            if BallTree is None:  # pragma: no cover
                raise ImportError("scikit-learn must be installed to search an unprojected graph")
            # haversine requires lat, lng coords in radians
            vertices_rad = np.deg2rad(vertices[["y", "x"]])
            points_rad = np.deg2rad(np.array([Y, X]).T)
            dist, pos = BallTree(vertices_rad, metric="haversine").query(points_rad, k=1)
            dist = dist[:, 0] * EARTH_RADIUS_M  # convert radians -> meters
            ne = vertices.index[pos[:, 0]]

    # convert results to correct types for return
    ne = list(ne)
    dist = list(dist)
    if is_scalar:
        ne = ne[0]
        dist = dist[0]

    if return_dist:
        return ne, dist
    else:
        return ne
    
def graph_to_gdf(G, nodes=False, edges=False, dual=False):

    crs = G.graph['crs']
    if nodes:
        if not G.nodes:  # pragma: no cover
            raise ValueError("graph contains no nodes")
            
        nodes, data = zip(*G.nodes(data=True))
        
        # convert node x/y attributes to Points for geometry column
        geom = (Point(d["x"], d["y"]) for d in data)
        gdf_nodes = gpd.GeoDataFrame(data, index=nodes, crs=crs, geometry=list(geom))
         
        if not dual:
            gdf_nodes.index.rename("osmid", inplace=True)
        
    if edges:
        if not G.edges:  # pragma: no cover
            raise ValueError("graph contains no edges")
        
        if not dual: 
            u, v, k, data = zip(*G.edges(keys=True, data=True))
        else: 
            u, v, data = zip(*G.edges(data=True))
            
        x_lookup = nx.get_node_attributes(G, "x")
        y_lookup = nx.get_node_attributes(G, "y")
        def make_geom(u, v, data, x= x_lookup, y= y_lookup):
            if "geometry" in data:
                return data["geometry"]
            else:
                return LineString((Point((x[u], y[u])), Point((x[v], y[v]))))
            
        geom = map(make_geom, u, v, data)
        gdf_edges = gpd.GeoDataFrame(data, crs=crs, geometry=list(geom))


        # add u, v, key attributes as index
        gdf_edges["u"] = u
        gdf_edges["v"] = v
        if not dual: 
            gdf_edges["key"] = k
            gdf_edges.set_index(["u", "v", "key"], inplace=True)
        else: 
            gdf_edges.set_index(["u", "v"], inplace=True)
    if nodes and edges:
        return gdf_nodes, gdf_edges
    elif nodes:
        return gdf_nodes
    elif edges:
        return gdf_edges
    
def project_gdf(gdf):
    mean_longitude = gdf["geometry"].representative_point().x.mean()

    # Compute UTM crs
    utm_zone = int(np.floor((mean_longitude + 180) / 6) + 1)
    utm_crs = f"+proj=utm +zone={utm_zone} +ellps=WGS84 +datum=WGS84 +units=m +no_defs"

    # project the GeoDataFrame to the UTM CRS
    gdf_proj = gdf.to_crs(utm_crs)
    # print(f"Projected to {gdf_proj.crs}")
    
    return gdf_proj

def check_align(bearing, back_bearing, angle, limit = 20):
    if (abs(angle - bearing) < limit) or (abs(angle - back_bearing) < limit):
        return True
    else:
        return False
    
def dissolve_poly(filepath, name):
    df = gpd.read_file(filepath)
    df['geometry'] = df.buffer(0.001)
    df[name] = name
    df = df.dissolve(name).reset_index()
    return df

def filter_by_datetime(df, timezone, start='9:00', end='18:00'):
    df['datetime'] = pd.to_datetime(df['time'], unit='ms')
    df = df.set_index('datetime')
    df.index = df.index.tz_localize("UTC").tz_convert(timezone)
    df = df.between_time(start, end)
    df = df.reset_index()
    df = df.drop(columns= ['time'])
    return df, len(df)

def filter_by_edge_length(proj_gdf, interval = 5):
    proj_gdf['expected_images'] = proj_gdf['length'].apply(lambda x: int(x/interval))
    proj_gdf = proj_gdf.drop_duplicates(subset='image_id', keep='first')
    proj_gdf = proj_gdf.groupby('edge_id').apply(sample_group, include_groups=False).reset_index(drop=True)
    return proj_gdf

def sample_group(group):
    # Check if the group size is less than expected_images
    if len(group) < group['expected_images'].iloc[0]:
        # If so, sample all rows
        return group.sample(len(group))
    else:
        # Otherwise, sample 'expected_images' rows
        return group.sample(group['expected_images'].iloc[0])

def download_csv_to_dataframe(url, chunk_size=1024):
    """
    Downloads a large CSV file from the given URL and reads it into a pandas DataFrame.
    
    Parameters:
    - url (str): URL of the CSV file to download.
    - chunk_size (int): Size of chunks to read from the response (default is 1024 bytes).
    
    Returns:
    - pd.DataFrame: DataFrame containing the CSV data.
    """
    # Stream the response
    response = requests.get(url, stream=True)
    
    # Check if the request was successful
    response.raise_for_status()
    
    # Create a StringIO object to hold the CSV data in memory
    csv_data = StringIO()
    
    # Stream the data and write it to the StringIO object
    for chunk in response.iter_content(chunk_size=chunk_size):
        if chunk:  # filter out keep-alive new chunks
            csv_data.write(chunk.decode('utf-8'))
    
    # Move to the start of the StringIO object
    csv_data.seek(0)
    
    # Read the CSV data into a pandas DataFrame
    df = pd.read_csv(csv_data)
    
    return df


def find_optimal_chunk_size(url, buffer_sizes=[8 * 1024, 16 * 1024, 32 * 1024, 64 * 1024]):
    """
    Finds the optimal chunk size for downloading a file from the given URL based on download time.
    
    Parameters:
    - url (str): URL of the file to download.
    - buffer_sizes (list of int): List of chunk sizes (in bytes) to test.
    
    Returns:
    - optimal_chunk_size (int): The buffer size that resulted in the shortest download time.
    - results (dict): Dictionary with buffer sizes as keys and times taken as values.
    """
    results = {}
    
    for buffer_size in buffer_sizes:
        print(f"Testing buffer size: {buffer_size} bytes")
        
        start_time = time.time()
        
        try:
            response = requests.get(url, stream=True)
            response.raise_for_status()
            
            # Temporarily store the file to measure download time
            with open('temp_download_file', 'wb') as file:
                for chunk in response.iter_content(chunk_size=buffer_size):
                    if chunk:
                        file.write(chunk)
        
        except Exception as e:
            print(f"Error occurred with buffer size {buffer_size}: {e}")
            results[buffer_size] = {'error': str(e)}
            continue
        
        end_time = time.time()
        elapsed_time = end_time - start_time
        
        results[buffer_size] = {'time': elapsed_time}
        print(f"Buffer size: {buffer_size} bytes, Time taken: {elapsed_time:.2f} seconds")
    
    # Determine the optimal chunk size
    optimal_chunk_size = min(results, key=lambda k: results[k]['time'] if 'time' in results[k] else float('inf'))
    
    return optimal_chunk_size, results

def get_image_url(image_id, session, api_key=''):
    """
    Query Mapillary Graph API for a single image's 256-thumb URL.
    """
    url = f'https://graph.mapillary.com/{image_id}?fields=thumb_256_url'
    resp = session.get(url, params={'access_token': api_key})
    resp.raise_for_status()  # Raises an HTTPError if the response was unsuccessful
    data = resp.json()
    return data.get('thumb_256_url', None)


def view_image(image_id, api_key = ''):

    url = f'https://graph.mapillary.com/{image_id}?fields=thumb_256_url'

    response = session.get(url, params={'access_token': api_key})
    image_data = response.json()
    image_url = image_data['thumb_256_url']

    response = session.get(image_url, stream=True)
    response = response.raw

    image_array = np.asarray(bytearray(response.read()), dtype=np.uint8)
    image = cv2.imdecode(image_array, cv2.IMREAD_COLOR)   
    
    # Convert BGR to RGB
    image_rgb = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

    return image_rgb

def draw_line(lon, lat, bearing):
    r = 0.0001  # or whatever fits you
    plt.arrow(lon, lat, r*math.cos(bearing), r*math.sin(bearing), width=0.0001, color='red') 

def save_image_to_folder(image_gdf, output_dir, session, api_key, max_retries=3, retry_delay=1):
    image_set = list(image_gdf['image_id'].unique())
    for image_id in image_set:
        image_path = f'{image_id}.jpg'
        url = f'https://graph.mapillary.com/{image_id}?fields=thumb_256_url'

        attempt = 0
        while attempt < max_retries:
            try:
                response = session.get(url, params={'access_token': api_key})
                response.raise_for_status()
                image_data = response.json()
                image_url = image_data['thumb_256_url']
        
                response = session.get(image_url, stream=True)
                response.raise_for_status()
                response = response.raw
            
                break  # Exit loop if request is successful
            except requests.RequestException as e:
                attempt += 1
                print(f'Connection error for image: {image_id} (Attempt {attempt}/{max_retries}): {e}')
                if attempt < max_retries:
                    time.sleep(retry_delay)  # Wait before retrying
                else:
                    print(f'Failed to fetch image: {image_id} after {max_retries} attempts.')
                    return []  # Return an empty list if all attempts fail

        image_array = np.asarray(bytearray(response.read()), dtype=np.uint8)
        image = cv2.imdecode(image_array, cv2.IMREAD_COLOR)   
        
        # Convert BGR to RGB
        image_rgb = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        cv2.imwrite(os.path.join(output_dir, image_path), image_rgb)

def assign_closest_points_to_linestring(image_gdf, street_gdf):
    svi_gdf = image_gdf.copy()
    # 0) Project to meters
    if svi_gdf.crs is None or svi_gdf.crs.is_geographic:
        points_gdf = svi_gdf.to_crs(3857)
        lines_gdf  = street_gdf.to_crs(3857)

    # 1) Build an STRtree over line geometries (cheap to query many times)
    line_geoms   = lines_gdf.geometry.values
    tree         = STRtree(line_geoms)
    line_ids_arr = lines_gdf.index.to_numpy()  # or use a dedicated id column

    # 2) Query all nearest lines at once
    pt_geoms = points_gdf.geometry.values
    # idx_pt: indices into pt_geoms; idx_ln: indices into line_geoms
    idx, dists = tree.query_nearest(pt_geoms, return_distance=True)
    idx_pt = idx[0,:]
    idx_ln = idx[1,:]
    # 3) Place results back into arrays aligned to points_gdf
    nearest_line_id = np.empty(len(points_gdf), dtype=line_ids_arr.dtype)
    nearest_dist_m  = np.empty(len(points_gdf), dtype="float64")
    nearest_line_id[idx_pt] = line_ids_arr[idx_ln]
    nearest_dist_m[idx_pt]  = dists

    # 4) Attach to a copy or the original GeoDataFrame
    out = points_gdf.copy()
    out["nearest_street_id"] = nearest_line_id
    out["dist_to_line"] = nearest_dist_m

    svi_gdf = svi_gdf.merge(out[['image_id', 'nearest_street_id', 'dist_to_line']], on='image_id')

    return svi_gdf

def assign_closest_points_to_linestring_and_sample(image_gdf, street_gdf, bin_distance = 20, proximity_distance = 10):
    svi_gdf = image_gdf.copy()
    if svi_gdf.crs is None or svi_gdf.crs.is_geographic:
        points_gdf = svi_gdf.to_crs(3857)
        lines_gdf  = street_gdf.to_crs(3857)

        # 1) Build STRtree on lines once
        line_geoms = lines_gdf.geometry.values
        tree = STRtree(line_geoms)

        # Use a stable line identifier
        line_ids_arr = lines_gdf.index.to_numpy()

        # 2) Nearest line for every point
        pt_geoms = points_gdf.geometry.values
        idx, dists = tree.query_nearest(pt_geoms, return_distance=True)
        idx_pt = idx[0,:]
        idx_ln = idx[1,:]

        # Align results to points_gdf’s index order
        n = len(points_gdf)
        nearest_line_idx = np.empty(n, dtype=np.int64)
        nearest_line_id = np.empty(n, dtype=line_ids_arr.dtype)
        nearest_dist_m = np.empty(n, dtype="float64")
        nearest_line_idx[idx_pt] = idx_ln
        nearest_line_id[idx_pt]  = line_ids_arr[idx_ln]
        nearest_dist_m[idx_pt] = dists

        # 3) Project each point onto its matched line to get arclength along the line
        #    This is O(N) in Python but uses shapely C-core per operation and is fast in practice
        proj_m = np.fromiter(
            (line_geoms[i].project(pt) for pt, i in zip(pt_geoms, nearest_line_idx)),
            dtype="float64", count=n
        )

        # 4) Bin by arclength along the line
        bins = (proj_m // bin_distance).astype(np.int64)

        # 5) Within each (line_id, bin) keep the point closest to the line (smallest distance)
        df = pd.DataFrame({
            "nearest_street_id": nearest_line_id,
            "bin": bins,
            "dist_m": nearest_dist_m,
        }, index=points_gdf.index)

        df = df[df['dist_m']<=proximity_distance]

        keep_idx = df.groupby(["nearest_street_id", "bin"])["dist_m"].idxmin()

        # 6) Return thinned points with metadata
        out = points_gdf.loc[keep_idx].copy()
        out["nearest_street_id"] = df.loc[keep_idx, "nearest_street_id"].values
        out["dist_to_line"]  = df.loc[keep_idx, "dist_m"].values
        out.attrs["bin_size_m"] = bin_distance
        svi_gdf = svi_gdf.merge(out[['image_id','nearest_street_id', 'dist_to_line']], on='image_id')

    return svi_gdf

def _default_class_colors():
    """Create a full color map for all classes, with sensible defaults for common labels."""
    num_classes = len(CLS_DICT_INT)
    cmap = cm.get_cmap('tab20', num_classes)
    class_colors = {cid: (np.array(cmap(i)[:3]) * 255).astype(int).tolist()
                    for i, cid in enumerate(sorted(CLS_DICT_INT.keys()))}
    # Preferred overrides
    class_colors[27] = [142, 202, 230]  # Sky
    class_colors[17] = [74, 87, 89]     # Building
    class_colors[30] = [96, 108, 56]    # Vegetation / Greenery
    class_colors[55] = [252, 163, 17]   # Car
    return class_colors

def _colorize_mask(mask: np.ndarray, class_colors: dict) -> np.ndarray:
    """Convert a 2D segmentation mask to an RGB image using class_colors {class_id: [r,g,b]}."""
    h, w = mask.shape
    rgb = np.zeros((h, w, 3), dtype=np.uint8)
    # Vectorized painting per class id
    for cid, color in class_colors.items():
        rgb[mask == cid] = color
    return rgb

def _compute_percentages(seg_mask: np.ndarray, ids_of_interest=None, min_pct=0.0, top_k=None):
    """
    Returns list of tuples (class_id, pct_float) sorted by descending percentage.
    ids_of_interest can be None for all present classes.
    min_pct filters out tiny classes.
    top_k limits the number of rows returned.
    """
    total = seg_mask.size
    unique, counts = np.unique(seg_mask, return_counts=True)
    pct_map = {int(k): 100.0 * c / total for k, c in zip(unique, counts)}

    # restrict to interest ids if provided
    items = pct_map.items()
    if ids_of_interest is not None:
        ids_set = set(int(i) for i in ids_of_interest)
        items = [(cid, pct) for cid, pct in items if cid in ids_set]

    # filter by min pct
    rows = [(cid, pct) for cid, pct in items if pct >= min_pct]

    # sort and cap
    rows.sort(key=lambda x: -x[1])
    if top_k is not None:
        rows = rows[:top_k]
    return rows

def _rgb255_to_hex(rgb):
    r, g, b = rgb
    return '#%02x%02x%02x' % (int(r), int(g), int(b))


def display_segmentation_grid(
    image_gdf,
    processor,
    model,
    device,
    rows=3,
    cols=4,
    id_col='image_id',
    mapillary_token_env='MAPILLARY_API_TOKEN',
    classes_to_color=['Sky', 'Building', 'Vegetation', 'Car'],       # list of labels for figure legend and also default ids_of_interest
    extra_color_overrides={'Sky': '#8ecae6',
                            'Building': '#4a5759',
                            'Vegetation': '#606c38',   # maps to Vegetation
                            'Car': '#ffc8dd',
                            },  # dict label -> hex or [r,g,b]
    title_fn=None,
    save_path=None,
    # New controls for per subplot percentage box
    per_subplot_min_pct=1.0,     # only show classes with percentage at least this value
    per_subplot_top_k=None,      # limit number of lines shown per subplot
    per_subplot_loc='lower left',# legend location within each segmentation axis
    per_subplot_fontsize=9,      # text size
    per_subplot_box=True         # add a small frame so text is readable
):
    assert cols % 2 == 0, "cols must be even since each image uses two columns"
    n_pairs = rows * (cols // 2)
    image_ids = list(image_gdf[id_col].values)[:n_pairs]

    # colors
    class_colors = _default_class_colors()

    def _to_rgb255(c):
        if isinstance(c, str) and c.startswith('#'):
            c = c.lstrip('#')
            return [int(c[i:i+2], 16) for i in (0, 2, 4)]
        if isinstance(c, (list, tuple)) and len(c) == 3:
            return list(map(int, c))
        raise ValueError("Color must be hex '#RRGGBB' or [r,g,b]")

    selected_labels = list(classes_to_color) if classes_to_color else []
    if extra_color_overrides:
        for label, color in extra_color_overrides.items():
            if label in LABEL_TO_ID_INT:
                class_colors[LABEL_TO_ID_INT[label]] = _to_rgb255(color)

    # figure level legend for selected labels
    legend_patches = []
    for label in selected_labels:
        if label in LABEL_TO_ID_INT:
            cid = LABEL_TO_ID_INT[label]
            legend_patches.append(
                mpatches.Patch(color=_rgb255_to_hex(class_colors[cid]), label=label)
            )

    fig_h = rows * 3.2
    fig_w = 16
    fig = plt.figure(figsize=(fig_w, fig_h))
    gs = GridSpec(rows, cols, figure=fig, hspace=0.02, wspace=0.05)

    token = os.environ.get(mapillary_token_env, None)
    if token is None:
        raise RuntimeError(f"Set {mapillary_token_env} in your environment")

    def _load_image(image_id):
        url = f'https://graph.mapillary.com/{image_id}?fields=thumb_2048_url'
        data = requests.get(url, params={'access_token': token}).json()
        im_url = data['thumb_2048_url']
        return Image.open(requests.get(im_url, stream=True).raw).convert('RGB')

    # ids of interest for percentage box
    ids_of_interest = None
    if selected_labels:
        ids_of_interest = [LABEL_TO_ID_INT[lbl] for lbl in selected_labels if lbl in LABEL_TO_ID_INT]

    for i, image_id in enumerate(image_ids):
        row = i // (cols // 2)
        col = (i % (cols // 2)) * 2

        im = _load_image(image_id)

        inputs = processor(images=im, return_tensors="pt")
        with torch.no_grad():
            pixel_values = inputs['pixel_values'].to(device)
            pixel_mask = inputs.get('pixel_mask')
            if pixel_mask is not None:
                outputs = model(pixel_values=pixel_values, pixel_mask=pixel_mask.to(device))
            else:
                outputs = model(pixel_values=pixel_values)

        target_size = [im.size[::-1]]
        seg = processor.post_process_semantic_segmentation(
            outputs, target_sizes=target_size
        )[0].to('cpu').numpy()
        seg_rgb = _colorize_mask(seg, class_colors)

        # original subplot
        ax0 = fig.add_subplot(gs[row, col])
        ax0.imshow(im)
        ax0.axis('off')
        if title_fn:
            ax0.set_title(str(title_fn(i, image_id, image_gdf.iloc[i] if i < len(image_gdf) else None)), fontsize=10)
        else:
            ax0.set_title(f'Original {i+1}', fontsize=10)

        # segmentation subplot
        ax1 = fig.add_subplot(gs[row, col + 1])
        ax1.imshow(seg_rgb, interpolation='nearest')
        ax1.axis('off')
        ax1.set_title('Segmented', fontsize=10)

        # compute and draw per subplot percentages
        rows_pct = _compute_percentages(
            seg_mask=seg,
            ids_of_interest=ids_of_interest,      # None means all classes present
            min_pct=per_subplot_min_pct,
            top_k=per_subplot_top_k
        )

        if rows_pct:
            handles = []
            for cid, pct in rows_pct:
                label = CLS_DICT_INT.get(cid, str(cid))
                color_hex = _rgb255_to_hex(class_colors.get(cid, [200, 200, 200]))
                handles.append(mpatches.Patch(color=color_hex, label=f"{label} {pct:.1f}%"))

            leg = ax1.legend(handles=handles,
                             loc=per_subplot_loc,
                             fontsize=per_subplot_fontsize,
                             frameon=per_subplot_box,
                             handlelength=0.8,
                             borderpad=0.3,
                             labelspacing=0.3)
            if per_subplot_box and leg is not None:
                leg.get_frame().set_alpha(0.75)

    # figure level legend
    if legend_patches:
        leg = fig.legend(handles=legend_patches, loc='upper center', handlelength=1,
                         bbox_to_anchor=(0.5, 0.93), ncol=min(len(legend_patches), 6),
                         frameon=False, prop={'size': 11})
        for patch in leg.get_patches():
            patch.set_height(10)
            patch.set_y(-0.5)

    plt.tight_layout()
    if save_path:
        plt.savefig(save_path, bbox_inches='tight', dpi=180)
    plt.show()


def fill_attributes_by_graph(
    gdf,
    aggr_cols,
    restrict_within_street=False,
    weight_col=None,
    inplace=False,
):

    if not inplace:
        gdf = gdf.copy()

    # Build MultiGraph
    G = nx.MultiGraph()
    for idx, row in gdf.iterrows():
        u, v = row["u"], row["v"]
        data = {"idx": idx}
        if weight_col is not None and (weight_col in gdf.columns) and pd.notna(row.get(weight_col)):
            data[weight_col] = row[weight_col]
        G.add_edge(u, v, key=idx, **data)

    def _incident_edge_indices(node):
        if node not in G:
            return []
        idxs = []
        for _, keyed in G[node].items():
            for _, d in keyed.items():
                idxs.append(d["idx"])
        return idxs

    def _nearest_seed_values(seed_values_per_node):
        import heapq
        dist = {n: np.inf for n in G.nodes}
        val = {n: np.nan for n in G.nodes}
        h = []
        for n, sv in seed_values_per_node.items():
            if not np.isnan(sv):
                dist[n] = 0.0
                val[n] = sv
                heapq.heappush(h, (0.0, n, sv))

        def _edge_weight(n, nbr):
            if weight_col is None:
                return 1.0
            weights = [d.get(weight_col, 1.0) for _, d in G[n][nbr].items()]
            return min(weights) if weights else 1.0

        visited = set()
        while h:
            d, n, sv = heapq.heappop(h)
            if n in visited:
                continue
            visited.add(n)
            dist[n] = d
            val[n] = sv
            for nbr in G.neighbors(n):
                if nbr in visited:
                    continue
                nd = d + _edge_weight(n, nbr)
                heapq.heappush(h, (nd, nbr, sv))
        return val, dist

    summary = {}
    street_groups = [("ALL", gdf.index)]
    if restrict_within_street:
        street_groups = list(gdf.groupby("street_id").groups.items())

    for col in aggr_cols:

        filled_count_total = 0
        for _, idxs in street_groups:
            rows_idx = pd.Index(idxs)

            nodes_in_group = set(gdf.loc[rows_idx, "u"]).union(set(gdf.loc[rows_idx, "v"]))

            node_seeds = {}
            for n in nodes_in_group:
                vals = []
                for eidx in _incident_edge_indices(n):
                    if eidx in rows_idx:
                        v = gdf.at[eidx, col]
                        if pd.notna(v):
                            vals.append(v)
                node_seeds[n] = np.mean(vals) if len(vals) else np.nan

            if not any(pd.notna(v) for v in node_seeds.values()):
                continue

            node_val, _ = _nearest_seed_values(node_seeds)

            # ✅ Correct row boolean indexing
            missing_mask = gdf.loc[rows_idx, col].isna()
            if not missing_mask.any():
                continue
            eidxs_to_fill = missing_mask.index[missing_mask.values]

            for eidx in eidxs_to_fill:
                u = gdf.at[eidx, "u"]
                v = gdf.at[eidx, "v"]
                candidates = [node_val.get(u, np.nan), node_val.get(v, np.nan)]
                candidates = [c for c in candidates if pd.notna(c)]
                if candidates:
                    gdf.at[eidx, col] = float(np.mean(candidates))
                    filled_count_total += 1


    return gdf

def fill_street_network(image_gdf, street_gdf):

    image_gdf_mod = image_gdf.copy()
    street_gdf_mod = street_gdf.copy()


    # Get target columns
    aggr_cols = ['nearest_street_id'] + list(image_gdf_mod.columns[9:])

    # Get mean of each street
    image_gdf_grouped = image_gdf_mod[aggr_cols].groupby(['nearest_street_id']).mean().reset_index()
    image_gdf_grouped = image_gdf_grouped.rename(columns={'nearest_street_id':'street_id'})

    # Merge to street
    street_gdf_mod = street_gdf_mod.merge(image_gdf_grouped, how='left', on = 'street_id')

    filled_gdf = fill_attributes_by_graph(street_gdf_mod,
                                        aggr_cols=aggr_cols[1:],
                                        restrict_within_street=False,  # set True to keep propagation within each street_id
                                        weight_col=None,               # or e.g. "length_m" if available
                                        inplace=False
                                        )
    return filled_gdf


import os
import io
import ee
import math
import json
import random
import glob
import geemap
import requests
import rasterio
import zipfile
import mercantile
import numpy as np
import pandas as pd
from PIL import Image
from tqdm import tqdm
import geopandas as gpd
from shapely import geometry
from shapely.geometry import Polygon, Point

from rasterio.mask import mask
from rasterio.io import MemoryFile
from rasterio.merge import merge
from rasterio.windows import Window
import matplotlib.pyplot as plt
import matplotlib.image as mpimg
import matplotlib.gridspec as gridspec
import rasterio.transform as rtransform
from shapely.geometry import box
from rasterio.transform import from_bounds
from rasterio.warp import calculate_default_transform, reproject, Resampling



def get_tiles_from_bbox(bbox: list = None,
                        zoom = 18):

    tiles = list(mercantile.tiles(bbox[0], bbox[1], bbox[2], bbox[3], zoom))
    
    return tiles

def download_satellite_tiles_from_bbox(bbox, api_key: str):
    # Get tiles from bounding box
    tiles = get_tiles_from_bbox(bbox) 
    
    # Create directory if not exists
    output_dir = "./satellite_data"
    os.makedirs(output_dir, exist_ok=True)
    
    # Initialize progress bar
    with tqdm(total=len(tiles), desc="Downloading Tiles") as pbar:
        for tile in tiles:
            img_path = os.path.join(output_dir, f"{tile.z}_{tile.x}_{tile.y}.jpg")
            
            # Check if file already exists
            if os.path.exists(img_path):
                pbar.update(1)
                continue
            
            tile_url = f'https://api.mapbox.com/v4/mapbox.satellite/{tile.z}/{tile.x}/{tile.y}@2x.jpg90?'
            response = requests.get(tile_url, params={'access_token': api_key})
            
            if response.status_code == 200:
                with open(img_path, 'wb') as f:
                    for chunk in response.iter_content(1024):
                        f.write(chunk)
            
            pbar.update(1)

def get_grid_size(tile_gdf_proj):
    x_grid_size = (tile_gdf_proj.iloc[[0]].geometry.total_bounds[2] - tile_gdf_proj.iloc[[0]].geometry.total_bounds[0])/512
    y_grid_size = (tile_gdf_proj.iloc[[0]].geometry.total_bounds[3] - tile_gdf_proj.iloc[[0]].geometry.total_bounds[1])/512
    return x_grid_size, y_grid_size

def get_max_img_dims(buildings_proj):
    ''' Helper function to return the maximum number of pixels in length and width to determine image chipset size'''
    
    bounds_df = buildings_proj.geometry.bounds
    x_max = max(bounds_df['maxx'] - bounds_df['minx'])
    y_max = max(bounds_df['maxy'] - bounds_df['miny'])
    total_max = max(x_max, y_max)
    half_max = total_max / 2
    return total_max, half_max


def get_and_combine_tiles(tiles_gdf_proj, building_chips, data_folder, output_folder):
    ''' Helper function to crop satellite image to desired building chip sizes'''
    
    # Placeholder list
    problems = []

    # Create directory to store building satellite image chips
    os.makedirs(os.path.join(os.getcwd(), output_folder), exist_ok=True)
        
    # Obtain chosen satellite images
    chosen_tiles = tiles_gdf_proj.overlay(building_chips)
    
    x_grid_size, y_grid_size = get_grid_size(tiles_gdf_proj)
    
    # Use tqdm to track progress
    for bid in tqdm(chosen_tiles['bid'].unique(), desc="Processing buildings"):
        temp_gdf = chosen_tiles[chosen_tiles['bid'] == bid]
        columns = []
        
        # First loop over x values
        for x_val in temp_gdf['X'].unique():
            x_gdf = temp_gdf[temp_gdf['X'] == x_val]
            
            rasters = []
            for _, row in x_gdf.iterrows():
                raster_path = os.path.join(os.getcwd(), data_folder, f"{row['Z']}_{row['X']}_{row['Y']}.jpg")
                with rasterio.open(raster_path) as f:
                    raster = f.read()
                    rasters.append(raster)
            
            columns.append(np.concatenate((rasters), axis=1))
            
        try:  
            # Combine satellite chips into single image
            img = np.concatenate((columns), axis=2)    
            # Extract building from image
            b_minx, b_miny, b_maxx, b_maxy = building_chips[building_chips['bid'] == bid]['geometry'].bounds.values[0]
            tile_minx, tile_miny, tile_maxx, tile_maxy = tiles_gdf_proj[tiles_gdf_proj['tile_id'].isin(list(chosen_tiles[chosen_tiles['bid'] == bid]['tile_id']))].total_bounds
            
            img_x_start = int((b_minx - tile_minx) / x_grid_size)
            img_y_start = int((tile_maxy - b_maxy) / y_grid_size)
            img_x_end = int((b_maxx - b_minx) / x_grid_size) + img_x_start
            img_y_end = img_y_start + int((b_maxy - b_miny) / y_grid_size)
            cropped_img = img[:, img_y_start:img_y_end, img_x_start:img_x_end]

            # Save image chip for building
            try:
                im = Image.fromarray(np.transpose(cropped_img, (1, 2, 0)))
                im.save(os.path.join(os.getcwd(), output_folder, f"{bid}.png"))
            except ValueError:
                problems.append(bid)
                
        except ValueError:
            problems.append(bid)
    
    # Try again for problems
    for bid in tqdm(problems, desc="Retrying failed cases"):
        temp_gdf = chosen_tiles[chosen_tiles['bid'] == bid]
        columns = []
        
        # First loop over x values
        for x_val in temp_gdf['X'].unique():
            rasters = []
            for y_val in temp_gdf['Y'].unique():
                raster_path = os.path.join(os.getcwd(), 'satellite_data/', f"18_{x_val}_{y_val}.jpg")
                with rasterio.open(raster_path) as f:
                    raster = f.read()
                    rasters.append(raster)
        
            columns.append(np.concatenate((rasters), axis=1))

        # Combine satellite chips into single image
        img = np.concatenate((columns), axis=2)    
        
        # Extract building from image
        b_minx, b_miny, b_maxx, b_maxy = building_chips[building_chips['bid'] == bid]['geometry'].bounds.values[0]
        tile_minx, tile_miny, tile_maxx, tile_maxy = tiles_gdf_proj[tiles_gdf_proj['tile_id'].isin(list(chosen_tiles[chosen_tiles['bid'] == bid]['tile_id']))].total_bounds

        img_x_start = int((b_minx - tile_minx) / x_grid_size)
        img_y_start = int((tile_maxy - b_maxy) / y_grid_size)
        img_x_end = int((b_maxx - b_minx) / x_grid_size) + img_x_start
        img_y_end = img_y_start + int((b_maxy - b_miny) / y_grid_size)
        cropped_img = img[:, img_y_start:img_y_end, img_x_start:img_x_end]

        # Save image chip for building
        im = Image.fromarray(np.transpose(cropped_img, (1, 2, 0)))
        im.save(os.path.join(os.getcwd(), output_folder, f"{bid}.png"))



def view_satellite_building_pair(gdf, bids):    
    # Initialize figure and a 4x4 GridSpec
    fig = plt.figure(figsize=(20, 20))
    gs = gridspec.GridSpec(4,4, wspace=0.3, hspace=0.3)
    
    chosen_bids = random.sample(sorted(bids), 8)
    for i, bid in enumerate(chosen_bids):
        # Determine row and column for this pair
        #
        # We have 4 rows total (0..3).
        # Each row has 4 columns (0..3).
        # Each row is split into 2 pairs: (col 0, col 1) and (col 2, col 3).
        # i // 2 is the row index, (i % 2)*2 is the column index for the polygon.
        row = i // 2
        col = (i % 2) * 2

        # 1) Plot the building polygon on the left subplot of the pair
        ax_poly = fig.add_subplot(gs[row, col])
        gdf[gdf['bid'] == bid].plot(ax=ax_poly, color='lightblue', edgecolor='black')
        ax_poly.set_title(f"BID: {bid} – Polygon")
        ax_poly.set_aspect('auto')
        ax_poly.set_xticks([])
        ax_poly.set_yticks([])

        # 2) Plot the corresponding satellite image on the right subplot
        ax_img = fig.add_subplot(gs[row, col+1])
        img_path = f'./building_satellite/{bid}.png'
        img = mpimg.imread(img_path)
        ax_img.imshow(img)
        ax_img.set_title(f"BID: {bid} – Satellite Image")
        ax_img.set_xticks([])
        ax_img.set_yticks([])
        ax_img.set_aspect('auto')

    plt.show()


def get_building_image_chips(building_proj, tiles_gdf_proj, add_context = False, pad_npixels = 0):    
    # Project building to local coordinates
    building_chips = building_proj.copy()
    
    if add_context:
        geom_list = []
        x_grid_size, y_grid_size = get_grid_size(tiles_gdf_proj)
        building_chips['geometry'] = building_chips['geometry'].bounds.apply(lambda row: geometry.box(row['minx']-x_grid_size*pad_npixels, 
                                                                                                      row['miny']-y_grid_size*pad_npixels, 
                                                                                                      row['maxx']+x_grid_size*pad_npixels, 
                                                                                                      row['maxy']+y_grid_size*pad_npixels), axis=1)
    else: 
        building_chips['geometry'] = building_chips['geometry'].bounds.apply(lambda row: geometry.box(row['minx'], row['miny'], row['maxx'], row['maxy']), axis=1)
    
    return building_chips


def get_tiles_gdf(tiles, bounds = None):
    x_list, y_list, z_list = [], [], []
    
    geom = []
    for tile in tiles:
        x_list.append(tile.x)
        y_list.append(tile.y)
        z_list.append(tile.z)
        geom.append(geometry.box(*mercantile.bounds(tile)))
    df = pd.DataFrame({'X': x_list, 'Y': y_list, 'Z': z_list})
    gdf = gpd.GeoDataFrame(data=df, crs = 'epsg:4326', geometry=geom)
    gdf['GID'] = range(len(gdf))
    
    if bounds is not None:
        intersect = bounds.overlay(gdf)
        selected_grids = gdf[gdf['GID'].isin(list(intersect['GID']))]
        selected_grids = selected_grids[['X','Y','Z','geometry']]

        return selected_grids
    
    return gdf


def download_tiff_from_path(lcz_path):
    # Send a request to get the file size (for progress calculation)
    response = requests.get(lcz_path, stream=True)
    
    if response.ok:
        total_size = int(response.headers.get('content-length', 0))  # Get total file size in bytes
        chunk_size = 1024 * 8  # Define chunk size
        save_path = './lcz_data/lcz_data.tif'

        # Open file and start downloading with progress bar
        with open(save_path, 'wb') as f, tqdm(
            desc="Downloading",
            total=total_size,
            unit='B',
            unit_scale=True,
            unit_divisor=1024
        ) as bar:
            for chunk in response.iter_content(chunk_size=chunk_size):
                if chunk:
                    f.write(chunk)
                    f.flush()
                    os.fsync(f.fileno())
                    bar.update(len(chunk))  # Update progress bar

        print("Downloaded global LCZ data to:", save_path)
    else:
        print(f"Failed to download file: {response.status_code}")

def reproject_raster(input_path, output_path):
    # Paths to input and output files


    # The target CRS is EPSG:4326
    dst_crs = "EPSG:4326"

    # Open the source dataset
    with rasterio.open(input_path) as src:
        # Calculate the transform and dimensions of the new raster
        transform, width, height = calculate_default_transform(
            src.crs,       # The current coordinate reference system
            dst_crs,       # The desired coordinate reference system
            src.width,
            src.height,
            *src.bounds     # The bounding coordinates of the original file
        )

        # Update metadata
        kwargs = src.meta.copy()
        kwargs.update({
            "crs": dst_crs,
            "transform": transform,
            "width": width,
            "height": height
        })

        # Create and write to the new reprojected raster
        with rasterio.open(output_path, "w", **kwargs) as dst:
            for i in range(1, src.count + 1):
                reproject(
                    source=rasterio.band(src, i),
                    destination=rasterio.band(dst, i),
                    src_transform=src.transform,
                    src_crs=src.crs,
                    dst_transform=transform,
                    dst_crs=dst_crs,
                    resampling=Resampling.nearest
                )

        print(f"Reprojected raster saved to {output_path}")
        
def raster2gdf(raster_path, chosen_band = 1, get_grid=True, zoom=False, boundary=None):
    """Helper function to convert a raster dataset (.tiff) to GeoDataFrame object.

    Args:
        raster_path (str): Filepath to a raster dataset (.tiff)
        chosen_band (int, optional): Selection of raster channel. Defaults to 1.
        get_grid (bool, optional): If True, returns gridded GeoDataFrame, otherwise returns center points. Defaults to True.
        zoom (bool, optional): If True, reads only a windowed portion of large .tiff file. Requires `boundary` to be provided. Defaults to False. 
        boundary (gpd.GeoDataFrame, optional): Geographic boundary to specify window extent for raster filtering. Defaults to None. 
    Returns:
        gpd.GeoDataFrame: A GeoDataFrame of points coordinates that correspond to original raster tile (centroid) positions.
    """    
    
    try:
        raster_dataset = rasterio.open(raster_path)
        boundary = boundary.to_crs(raster_dataset.crs)
    except:
        with rasterio.open(raster_path) as src:
            print(src.profile)

    if zoom and boundary is not None:

        xarray = np.linspace(raster_dataset.bounds[0], raster_dataset.bounds[2],raster_dataset.width+1)
        yarray = np.linspace(raster_dataset.bounds[3], raster_dataset.bounds[1],raster_dataset.height+1)
        half_x_grid_size = (raster_dataset.bounds[2] - raster_dataset.bounds[0])/(raster_dataset.width)/2
        half_y_grid_size = (raster_dataset.bounds[3] - raster_dataset.bounds[1])/(raster_dataset.height)/2

        # Get boundary coordinates
        xmin, ymin, xmax, ymax = boundary.total_bounds

        # Get raster offsets, width, and height
        list_x_range = [i for i in list(xarray) if i >= xmin-half_x_grid_size if i <=xmax+half_x_grid_size]
        list_y_range = [i for i in list(yarray) if i >= ymin-half_y_grid_size if i <=ymax+half_y_grid_size]
        x_width = len(list_x_range)
        y_width = len(list_y_range)
        x_off = np.where(xarray == list_x_range[0])[0].item()
        y_off = np.where(yarray == list_y_range[0])[0].item()

        geom_list = []
        for y_i in range(len(list_y_range)-1):
            for x_i in range(len(list_x_range)-1): 
                geom_list.append(Polygon([(list_x_range[x_i], list_y_range[y_i]), 
                                        (list_x_range[x_i], list_y_range[y_i+1]), 
                                        (list_x_range[x_i+1], list_y_range[y_i+1]), 
                                        (list_x_range[x_i+1], list_y_range[y_i])]))

        # Read windowed raster view
        with rasterio.open(raster_path) as src:
            value_array = src.read(1, window=Window(x_off, y_off, x_width-1, y_width-1))
        
        mydf = pd.DataFrame({'value':np.ravel(value_array, order='C')})
        my_gdf = gpd.GeoDataFrame(data = mydf, crs = raster_dataset.crs, geometry = geom_list)
        my_gdf = my_gdf.to_crs('epsg:4326')
        return my_gdf

    # Raster value array
    value_array = raster_dataset.read(chosen_band)
    
    if get_grid:
        xarray = np.linspace(raster_dataset.bounds[0], raster_dataset.bounds[2],raster_dataset.width+1)
        yarray = np.linspace(raster_dataset.bounds[3], raster_dataset.bounds[1],raster_dataset.height+1)
        
        mydf = pd.DataFrame({'value':np.ravel(value_array, order='C')})
        geom_list = []
        for y_i in range(len(yarray)-1):
            for x_i in range(len(xarray)-1): 
                geom_list.append(Polygon([(xarray[x_i], yarray[y_i]), (xarray[x_i], yarray[y_i+1]), (xarray[x_i+1], yarray[y_i+1]), (xarray[x_i+1], yarray[y_i])]))
        gdf = gpd.GeoDataFrame(data = mydf, crs = raster_dataset.crs, geometry = geom_list)
    else:
        # Get longitudinal half-grid size
        half_x_grid_size = (raster_dataset.bounds[2] - raster_dataset.bounds[0])/(raster_dataset.width)/2
        half_y_grid_size = (raster_dataset.bounds[3] - raster_dataset.bounds[1])/(raster_dataset.height)/2

        # Get longitudinal intervals
        xarray = np.linspace(raster_dataset.bounds[0]+half_x_grid_size, raster_dataset.bounds[2]-half_x_grid_size,raster_dataset.width)
        yarray = np.linspace(raster_dataset.bounds[3]-half_y_grid_size, raster_dataset.bounds[1]+half_y_grid_size,raster_dataset.height)

        # Obtain full list of coordinates
        lat_coords = np.repeat(yarray,raster_dataset.width)
        long_coords = np.tile(xarray, raster_dataset.height)

        # Option C - Changing from last index dimension first 
        gdf = gpd.GeoDataFrame(data = {'value':np.ravel(value_array, order='C')}, crs = raster_dataset.crs, geometry = gpd.points_from_xy(x=long_coords, y=lat_coords))
        gdf = gdf.to_crs('epsg:4326')
    return gdf

def raster2gdf_with_reprojection(raster_path, boundary_gdf, chosen_band=1):
    """
    Example function that:
      1) Reads a raster,
      2) Reprojects entire data to EPSG:4326 in memory,
      3) Zooms to a user boundary, also in EPSG:4326,
      4) Returns a polygonal GeoDataFrame of the zoomed portion.
    """
    dst_crs = "EPSG:4326"
    # ---------------------------
    # 1. Read entire source raster
    # ---------------------------
    with rasterio.open(raster_path) as src:
        src_data = src.read(chosen_band)
        src_transform = src.transform
        src_crs = src.crs
        width = src.width
        height = src.height
        bounds = src.bounds

    # ---------------------------
    # 2. Reproject to EPSG:4326
    # ---------------------------

    
    dst_transform, dst_width, dst_height = calculate_default_transform(
        src_crs,        # from
        dst_crs,        # to
        width, 
        height, 
        *bounds
    )

    # Prepare empty array for reprojected data
    dst_data = np.empty((dst_height, dst_width), dtype=src_data.dtype)

    reproject(
        source=src_data,
        destination=dst_data,
        src_transform=src_transform,
        src_crs=src_crs,
        dst_transform=dst_transform,
        dst_crs=dst_crs,
        resampling=Resampling.nearest
    )

    # Reproject the boundary to EPSG:4326 as well
    boundary_gdf = boundary_gdf.to_crs(dst_crs)
    # ---------------------------
    # 3. Build the full coordinate mesh (EPSG:4326)
    #    for the *entire* reprojected raster
    # ---------------------------
    # array_bounds returns (bottom, top) in final elements for y, so be mindful
    full_left, full_bottom, full_right, full_top = array_bounds(
        dst_height, dst_width, dst_transform
    )

    xarray_full = np.linspace(full_left,   full_right,  dst_width + 1)
    yarray_full = np.linspace(full_top,    full_bottom, dst_height + 1)

    # Pixel half-sizes in the reprojected space
    half_x = (full_right  - full_left)   / dst_width  / 2
    half_y = (full_top    - full_bottom) / dst_height / 2

    # ---------------------------
    # 4. Determine the zoom window in EPSG:4326
    #    from boundary_gdf's total_bounds
    # ---------------------------
    xmin, ymin, xmax, ymax = boundary_gdf.total_bounds

    # We want to include any pixel whose boundary overlaps
    # the boundary_gdf bounding box
    list_x_range = [
        x for x in xarray_full
        if (x >= xmin - half_x and x <= xmax + half_x)
    ]
    list_y_range = [
        y for y in yarray_full
        if (y <= ymax + half_y and y >= ymin - half_y)
    ]

    # Handle edge cases (e.g. boundary outside the raster)
    if not list_x_range or not list_y_range:
        print("Boundary is entirely outside the reprojected raster.")
        return None

    # Offsets in the reprojected array
    x_off = np.where(xarray_full == list_x_range[0])[0].item()
    y_off = np.where(yarray_full == list_y_range[0])[0].item()

    # The width/height in pixels of the zoomed region
    x_width = len(list_x_range) - 1
    y_height = len(list_y_range) - 1

    # ---------------------------
    # 5. Slice the reprojected data to get only the zoom
    #    region
    # ---------------------------
    # The data array has shape = (dst_height, dst_width)
    zoomed_data = dst_data[y_off:(y_off + y_height), x_off:(x_off + x_width)]

    # Build a new transform for this subset
    from rasterio.windows import Window, transform
    window = Window(x_off, y_off, x_width, y_height)
    zoom_transform = transform(window, dst_transform)

    # ---------------------------
    # 6. Create polygons for each pixel in the zoom region
    # ---------------------------
    zoom_left, zoom_bottom, zoom_right, zoom_top = rasterio.transform.array_bounds(
        y_height, x_width, zoom_transform
    )
    # Note: array_bounds -> (left, bottom, right, top)

    zoom_xarray = np.linspace(zoom_left,  zoom_right,  x_width  + 1)
    zoom_yarray = np.linspace(zoom_top,   zoom_bottom, y_height + 1)

    # Flatten the pixel values
    vals = zoomed_data.ravel(order='C')
    df = pd.DataFrame({'value': vals})

    polygons = []
    for row_i in range(y_height):
        for col_i in range(x_width):
            p = Polygon([
                (zoom_xarray[col_i],   zoom_yarray[row_i]),
                (zoom_xarray[col_i],   zoom_yarray[row_i+1]),
                (zoom_xarray[col_i+1], zoom_yarray[row_i+1]),
                (zoom_xarray[col_i+1], zoom_yarray[row_i])
            ])
            polygons.append(p)

    gdf = gpd.GeoDataFrame(df, geometry=polygons, crs=dst_crs)
    return gdf

def mosaic_and_save_tiff(source, dest, name, fmt='tiff'):
    tifs = glob.glob(os.path.join(source,  '*'))

    src_files_to_mosaic = []
    for fp in tifs:
        src = rasterio.open(fp)
        src_files_to_mosaic.append(src)
        
    mosaic, out_trans = merge(src_files_to_mosaic)

    if fmt == 'tiff':
        out_meta = src.meta
        out_meta.update({"driver": "GTiff",
                             "height": mosaic.shape[1],
                             "width": mosaic.shape[2],
                             "transform": out_trans})
        
        with rasterio.open(os.path.join(dest,f'{name}.tif'), "w", **out_meta) as dest:
            dest.write(mosaic)
            
    elif fmt == 'npz':
        out_meta = src.meta
        out_meta.update({
            "driver": "GTiff",
            "height": mosaic.shape[1],
            "width": mosaic.shape[2],
            "transform": out_trans
        })
        np.savez_compressed(os.path.join(dest, f'{name}.npz'), 
                            mosaic=mosaic, 
                            meta=out_meta)
        

def extract_tiff_from_shapefile(geotiff_path, shapefile, output_filepath=None, zipped=True):

    if zipped:
        with zipfile.ZipFile(geotiff_path, 'r') as zip_ref:
            zip_ref.extractall(os.path.dirname(geotiff_path))
        geotiff_path = geotiff_path[:-4]

    # Get geotiff info and mask with shapefile
    with rasterio.open(geotiff_path, 'r') as data:
        out_image, out_transform = rasterio.mask.mask(data, shapes=[shapefile.geometry[0]], crop=True)
        out_meta = data.meta
  

    # Set transformation of geotiff
    out_meta.update({"driver": "GTiff",
                     "height": out_image.shape[1],
                     "width": out_image.shape[2],
                     "transform": out_transform})
    
    # If output_filepath, write to disk
    if output_filepath:
        with rasterio.open(output_filepath, "w", **out_meta) as dest:
            dest.write(out_image)
    else:
        with MemoryFile() as memfile:
            with memfile.open(**out_meta) as dataset:
                dataset.write(out_image)
                
                # Read the data from the in-memory dataset
                data_array = dataset.read(1)
                data_array = np.nan_to_num(data_array, nan=0)
                
        return data_array
    

def merge_raster_to_gdf(raster, gdf, id_col = 'plot_id', raster_col='value', num_classes = 1, method='mean'):
    res_intersection = raster.overlay(gdf)

    if method == 'proportion':
        val_series = res_intersection.groupby(id_col)[raster_col].value_counts()

        raster_df = pd.DataFrame(index = val_series.index, data = val_series.values).reset_index()

        # return lcz_df
        raster_df = pd.pivot(raster_df, index=id_col, columns=raster_col, values=0).fillna(0)
    
        # Sum across rows to get total for each row
        row_totals = raster_df.sum(axis=1)
        raster_df = raster_df.div(row_totals, axis=0) * 100

        # Rename columns
        raster_df.columns = [raster_col + '_' + str(col) for col in raster_df.columns]
        raster_column_names = [raster_col + '_' +str(i) for i in range(0,num_classes)]

        for i in raster_column_names:
            if i not in set(raster_df.columns):
                raster_df[i] = 0
            elif i in set(raster_df.columns):
                raster_df[i] = raster_df[i].replace(np.nan, 0)
        
        raster_df = raster_df[raster_column_names]
        
        # Join pixel mean to network edges
        gdf = gdf.merge(raster_df, on=id_col, how='left')

    elif method == 'mean': 
        mean_series = res_intersection.groupby(id_col)[raster_col].mean()
        mean_series.name = f'{raster_col}_mean'

        std_series = res_intersection.groupby(id_col)[raster_col].std()
        std_series.name = f'{raster_col}_std'
        
        # Join pixel mean to network edges
        gdf = gdf.merge(mean_series, on=id_col, how='left')
        gdf[f'{raster_col}_mean'] = gdf[f'{raster_col}_mean'].fillna(0)

        gdf = gdf.merge(std_series, on=id_col, how='left')
        gdf[f'{raster_col}_std'] = gdf[f'{raster_col}_std'].fillna(0)

    return gdf

def tiled_large_array(boundary, layer_name, delta=0.1, band='', index=0):
    """Utility function to return mosaic array for large area by first tiling it. 

    Args:
        boundary (GeoDataFrame): The boundary polygon(s) for the area of interest.
        layer_name (str): Earth Engine ImageCollection ID or Image ID.
        delta (float): Optional parameter (not strictly needed if using ee_to_numpy).
        band (str): The band name to select. If provided, we select by band name.
        index (int): The index of the image in the collection to fetch (if multiple images).

    Returns:
        Full numpy array with mosaic-ed raster tiles
    """
    array_list = []

    tiles = split_bbox_into_grid(boundary, n=64)

    for i in range(len(tiles)):
        current_row = tiles.iloc[[i]]

        # 2. Convert boundary to Earth Engine geometry
        tile_geom = current_row[['geometry']]  # Keep only geometry column
        gdf_json = json.loads(tile_geom.to_json())
        ee_fc = ee.FeatureCollection(gdf_json["features"])
        roi = ee_fc.geometry()

            # 3. Attempt to load layer_name as an ImageCollection and pick one by index.
        #    If that fails or yields an empty collection, treat as single Image.
        try:
            # Try as an ImageCollection
            collection = ee.ImageCollection(layer_name).filterBounds(roi)
            count = collection.size().getInfo()
            if count > 0:
                # Sort by time and select the desired image
                sorted_collection = collection.sort("system:time_start", False)
                if index >= count:
                    raise ValueError(
                        f"Requested index {index} is out of range. Only {count} images found in collection."
                    )
                image_ee = ee.Image(sorted_collection.toList(count).get(index))
            else:
                # Fallback to treating layer_name as single Image
                image_ee = ee.Image(layer_name).clip(roi)
        except Exception:
            # If it fails as a collection, interpret as a single Image
            image_ee = ee.Image(layer_name).clip(roi)

        # Check available bands
        info = image_ee.getInfo()
        available_bands = [b['id'] for b in info['bands']]

        # 4. If the user specified a band name, select it; otherwise select the first band
        if band != '':
            if band not in available_bands:
                raise ValueError(
                    f"Error: '{band}' is not a valid band. Available bands are: {available_bands}"
                )
            image_ee = image_ee.select(band).clip(roi).unmask()
        else:
            first_band_name = available_bands[0]
            image_ee = image_ee.select(first_band_name).clip(roi).unmask()
        
        arr = geemap.ee_to_numpy(image_ee, region=roi)
        array_list.append(arr)

    return concat_grid(array_list)
        
def gee_layer_from_boundary(boundary, layer_name, delta=0.1, band='', index=0, large=True):
    """
    Fetch a raster from the given Earth Engine ImageCollection or single Image 
    over the input boundary, and convert it to a GeoDataFrame with polygon 
    geometries for each pixel.

    Args:
        boundary (GeoDataFrame): The boundary polygon(s) for the area of interest.
        layer_name (str): Earth Engine ImageCollection ID or Image ID.
        delta (float): Optional parameter (not strictly needed if using ee_to_numpy).
        band (str): The band name to select. If provided, we select by band name.
        index (int): The index of the image in the collection to fetch (if multiple images).

    Returns:
        GeoDataFrame with columns ['value'] and the polygon geometry of each pixel.
    """
    # Toggle tiling mechanism if area is large and want to use large
    if (boundary.geometry.area.values > 0.1) and large:
        large=True
    else:
        large=False
    
    if large:
        # 2. Convert boundary to Earth Engine geometry
        boundary = boundary[['geometry']]  # Keep only geometry column
        gdf_json = json.loads(boundary.to_json())
        ee_fc = ee.FeatureCollection(gdf_json["features"])
        roi = ee_fc.geometry()

        # 3. Attempt to load layer_name as an ImageCollection and pick one by index.
        #    If that fails or yields an empty collection, treat as single Image.
        try:
            # Try as an ImageCollection
            collection = ee.ImageCollection(layer_name).filterBounds(roi)
            count = collection.size().getInfo()
            if count > 0:
                # Sort by time and select the desired image
                sorted_collection = collection.sort("system:time_start", False)
                if index >= count:
                    raise ValueError(
                        f"Requested index {index} is out of range. Only {count} images found in collection."
                    )
                image_ee = ee.Image(sorted_collection.toList(count).get(index))
            else:
                # Fallback to treating layer_name as single Image
                image_ee = ee.Image(layer_name).clip(roi)
        except Exception:
            # If it fails as a collection, interpret as a single Image
            image_ee = ee.Image(layer_name).clip(roi)

        # Check available bands
        info = image_ee.getInfo()
        available_bands = [b['id'] for b in info['bands']]

        # 4. If the user specified a band name, select it; otherwise select the first band
        if band != '':
            if band not in available_bands:
                raise ValueError(
                    f"Error: '{band}' is not a valid band. Available bands are: {available_bands}"
                )
            image_ee = image_ee.select(band).clip(roi).unmask()
        else:
            first_band_name = available_bands[0]
            image_ee = image_ee.select(first_band_name).clip(roi).unmask()

        arr = tiled_large_array(boundary, layer_name, delta=0.1, band='', index=0)
        return arr

    else:

        # 2. Convert boundary to Earth Engine geometry
        boundary = boundary[['geometry']]  # Keep only geometry column
        gdf_json = json.loads(boundary.to_json())
        ee_fc = ee.FeatureCollection(gdf_json["features"])
        roi = ee_fc.geometry()

        # 3. Attempt to load layer_name as an ImageCollection and pick one by index.
        #    If that fails or yields an empty collection, treat as single Image.
        try:
            # Try as an ImageCollection
            collection = ee.ImageCollection(layer_name).filterBounds(roi)
            count = collection.size().getInfo()
            if count > 0:
                # Sort by time and select the desired image
                sorted_collection = collection.sort("system:time_start", False)
                if index >= count:
                    raise ValueError(
                        f"Requested index {index} is out of range. Only {count} images found in collection."
                    )
                image_ee = ee.Image(sorted_collection.toList(count).get(index))
            else:
                # Fallback to treating layer_name as single Image
                image_ee = ee.Image(layer_name).clip(roi)
        except Exception:
            # If it fails as a collection, interpret as a single Image
            image_ee = ee.Image(layer_name).clip(roi)

        # Check available bands
        info = image_ee.getInfo()
        available_bands = [b['id'] for b in info['bands']]

        # 4. If the user specified a band name, select it; otherwise select the first band
        if band != '':
            if band not in available_bands:
                raise ValueError(
                    f"Error: '{band}' is not a valid band. Available bands are: {available_bands}"
                )
            image_ee = image_ee.select(band).clip(roi).unmask()
        else:
            first_band_name = available_bands[0]
            image_ee = image_ee.select(first_band_name).clip(roi).unmask()

        # 5. Convert to NumPy array
        # arr = fetch_tiled_array(image_ee, roi, max_tile_px=2048)
        arr = geemap.ee_to_numpy(image_ee, region=roi)

    if arr is None:
        raise ValueError(
            "ee_to_numpy returned None. Possibly no valid pixels in the region or an error retrieving data."
        )

    # Handle single- or multi-band arrays
    if arr.ndim == 3:
        arr_2d = arr[:, :, 0]
    else:
        arr_2d = arr

    nrows, ncols = arr_2d.shape

    # 6. Get the bounding box from the ROI
    bounds_info = roi.bounds().getInfo()['coordinates'][0]
    minx = min(coord[0] for coord in bounds_info)
    maxx = max(coord[0] for coord in bounds_info)
    miny = min(coord[1] for coord in bounds_info)
    maxy = max(coord[1] for coord in bounds_info)

    # 7. Get the projection info (CRS) from the Image
    proj = image_ee.projection()
    proj_info = proj.getInfo()

    if 'crs' in proj_info:
        crs = proj_info['crs']
    else:
        # Fallback or guess. Often EPSG:4326 is a reasonable default
        crs = "EPSG:4326"

    # 8. Create an affine transform from bounding box and array shape
    transform = rtransform.from_bounds(minx, miny, maxx, maxy, ncols, nrows)

    # 9. Build polygons for each pixel
    polygons = []
    values = []

    for row in range(nrows):
        for col in range(ncols):
            pixel_value = arr_2d[row, col]

            # Top-left corner of pixel
            xleft, ytop = transform * (col, row)
            # Bottom-right corner of pixel
            xright, ybottom = transform * (col + 1, row + 1)

            poly = Polygon([
                (xleft, ytop),
                (xright, ytop),
                (xright, ybottom),
                (xleft, ybottom),
                (xleft, ytop)  # close polygon
            ])

            polygons.append(poly)
            values.append(pixel_value)
            
    # points = []
    # values = []

    # for row in range(nrows):
    #     for col in range(ncols):
    #         pixel_value = arr_2d[row, col]

    #         # (x_center, y_center) is the center of the pixel
    #         x_center, y_center = transform * (col + 0.5, row + 0.5)

    #         point = Point(x_center, y_center)
    #         points.append(point)
    #         values.append(pixel_value)


    # 10. Create a GeoDataFrame
    df = pd.DataFrame({"value": values})
    gdf = gpd.GeoDataFrame(df, geometry=polygons, crs=crs)

    return gdf

def concat_grid(tile_list, *, nrows=None, ncols=None, fill_value=np.nan):
    """
    Stitch a flat list of NumPy arrays—supplied row-major—into one mosaic.

    Parameters
    ----------
    tile_list : list[np.ndarray]
        Tiles in row-major order:  [row-0-col-0, row-0-col-1, …, row-1-col-0, …]
    nrows, ncols : int, optional
        Grid dimensions.  • If both are None, they are inferred by
          *sqrt*—works for 4, 16, 64, … (perfect squares).
        • If only one is given, the other is derived from `len(tile_list)`.
    fill_value : scalar, default np.nan
        Value used when padding tiles whose shapes differ within a row/column.

    Returns
    -------
    np.ndarray
        The stitched raster.

    Raises
    ------
    ValueError
        If the list length is incompatible with the requested grid.
    """
    n_tiles = len(tile_list)

    # ---- 1. derive grid shape ----------------------------------------
    if nrows is None and ncols is None:
        root = int(math.isqrt(n_tiles))
        if root * root != n_tiles:
            raise ValueError("tile_list length is not a perfect square; "
                             "specify nrows and/or ncols.")
        nrows = ncols = root
    elif nrows is None:
        if n_tiles % ncols:
            raise ValueError("ncols does not divide tile count.")
        nrows = n_tiles // ncols
    elif ncols is None:
        if n_tiles % nrows:
            raise ValueError("nrows does not divide tile count.")
        ncols = n_tiles // nrows
    else:
        if nrows * ncols != n_tiles:
            raise ValueError("nrows × ncols != tile count.")

    # ---- 2. find max height per row & max width per col --------------
    max_h_per_row = [
        max(tile_list[r * ncols + c].shape[0] for c in range(ncols))
        for r in range(nrows)
    ]
    max_w_per_col = [
        max(tile_list[r * ncols + c].shape[1] for r in range(nrows))
        for c in range(ncols)
    ]

    # ---- 3. pad tiles so shapes align within grid --------------------
    padded = []
    for r in range(nrows):
        for c in range(ncols):
            t = tile_list[r * ncols + c]
            h_pad = max_h_per_row[r] - t.shape[0]
            w_pad = max_w_per_col[c] - t.shape[1]
            if h_pad or w_pad:
                t = np.pad(
                    t,
                    ((0, h_pad), (0, w_pad)),
                    mode="constant",
                    constant_values=fill_value,
                )
            padded.append(t)

    # ---- 4. horizontal then vertical concat --------------------------
    rows = [
        np.hstack(padded[r * ncols : (r + 1) * ncols]) for r in range(nrows)
    ]
    mosaic = np.vstack(rows)
    return mosaic

def split_bbox_into_grid(bbox_gdf, *, n=None, nrows=None, ncols=None,
                         id_col="tile_id", row_major=True):
    """
    Split the bounding box of `bbox_gdf` into `nrows × ncols` equal-sized
    rectangles, or into *≈ n* rectangles if only `n` is given.

    Parameters
    ----------
    bbox_gdf : GeoDataFrame
        Any GeoDataFrame whose `.total_bounds` defines the area to split.
    n : int, optional
        Desired number of tiles (e.g. 4, 8, 16).  
        If `nrows`/`ncols` are **not** supplied, the function chooses a
        rectangular grid whose tile count ≥ `n` and is as square as possible.
    nrows, ncols : int, optional
        Explicit grid dimensions.  If one is `None`, it is derived from
        the other and `n` (e.g. `nrows=2, n=None, ncols=None` → `ncols`
        chosen so `2×ncols ≥ n`).
    id_col : str, default "tile_id"
        Name of the identifier column in the returned GeoDataFrame.
    row_major : bool, default True
        If `True`, IDs increase left-to-right, top-to-bottom (matrix style).
        If `False`, they increase column-wise (top-to-bottom, then left-right).

    Returns
    -------
    GeoDataFrame
        Grid cells in the same CRS as the input.  Extra cells (if the chosen
        grid has more than `n` tiles) are trimmed from the **bottom-right**
        corner so the row/column structure stays intact.
    """
    # ---- derive grid shape -------------------------------------------
    if nrows is None and ncols is None:
        if n is None:
            raise ValueError("Specify `n` or (`nrows`, `ncols`).")
        # choose the most square grid with rows*cols >= n
        nrows = int(math.floor(math.sqrt(n)))
        ncols = int(math.ceil(n / nrows))
    elif nrows is None:                    # ncols is given
        if n is None:
            raise ValueError("Need `n` to derive nrows.")
        nrows = int(math.ceil(n / ncols))
    elif ncols is None:                    # nrows is given
        if n is None:
            raise ValueError("Need `n` to derive ncols.")
        ncols = int(math.ceil(n / nrows))

    # ---- bounds -------------------------------------------------------
    minx, miny, maxx, maxy = bbox_gdf.total_bounds
    dx = (maxx - minx) / ncols
    dy = (maxy - miny) / nrows

    # ---- build tiles --------------------------------------------------
    tiles, ids = [], []
    current_id = 0
    for r in range(nrows):
        y_top    = maxy - r * dy
        y_bottom = maxy - (r + 1) * dy
        for c in range(ncols):
            if n is not None and current_id >= n:          # trim extras
                break
            x_left  = minx + c * dx
            x_right = minx + (c + 1) * dx
            tiles.append(box(x_left, y_bottom, x_right, y_top))
            ids.append(current_id)
            current_id += 1
        if n is not None and current_id >= n:
            break

    # ---- reorder IDs (row-major vs column-major) ----------------------
    if not row_major:
        # transpose indexing: id = c * nrows + r  instead of r * ncols + c
        ids = sorted(range(len(tiles)),
                     key=lambda k: (k % nrows) * ncols + (k // nrows))

    return gpd.GeoDataFrame({id_col: ids},
                            geometry=tiles,
                            crs=bbox_gdf.crs)


def get_affine_transform(gdf, raster):

    # ❶ Bounding box (in the same CRS you want for the raster)
    minx, miny, maxx, maxy = gdf.total_bounds   # (west, south, east, north)

    # ❷ Raster dimensions
    if raster.ndim == 3:          # (bands, rows, cols)
        height, width = raster[:,:,0].shape
    else:                      # (rows, cols)
        height, width = raster.shape

    transform = from_bounds(
        minx, miny, maxx, maxy,   # geographic extent
        width, height             # raster grid size
    )

    return transform


# import external packages
import networkx as nx
import numpy as np
import pandas as pd
import networkit

def compute_centrality(G, G_nodes, function, colname, T_nodes, *args, temporal=False):
    """Employs networkit fast centrality computation to obtain various network centrality measures.

    Args:
        G (nx.MultiDiGraph): Urban network graph object with nodes and edges information. 
        G_nodes (gpd.GeoDataFrame): Geopandas dataframe object with node attributes. 
        function (networkit.centrality.Centrality): Centrality function from networkit.centrality module. Accepts (e.g., EigenvectorCentrality, KatzCentrality, BetweennessCentrality).
        colname (str): New column to add computed centrality indicators to node attribute dataframe.

    Returns:
        gpd.GeoDataFrame: Node attribute dataframe with network centrality measures included. 
    """    
    G_nodes_mod = G_nodes.copy()
    
    nk_G = networkit.nxadapter.nx2nk(G, weightAttr=None)
    nk_centrality = function(nk_G, *args)
    nk_centrality.run()

    value = [v for k,v in nk_centrality.ranking()[:]]
    if temporal:
        T_nodes_mod= T_nodes.copy()
        T_nodes_mod[colname] = value
        T_nodes_mod = T_nodes_mod[['osmid', colname]]
        G_nodes_mod = pd.merge(G_nodes_mod, T_nodes_mod, on='osmid', how='left')
        G_nodes_mod[colname] = G_nodes_mod[colname].fillna(0)
        G_nodes_mod[colname] = G_nodes_mod[colname].round(3)
    else:
        
        G_nodes_mod[colname] = value
        G_nodes_mod[colname] = G_nodes_mod[colname].round(3)
    
    return G_nodes_mod

def merge_nx_attr(G, G_nodes, nxfunction, colname, T_nodes, temporal=False ,**kwargs):
    """Add graph attributes from networkx graph into node attribute dataframe. 

    Args:
        G (nx.MultiDiGraph): Urban network graph with nodes and edges information. 
        G_nodes (gpd.GeoDataFrame): Geopandas dataframe object with node attributes. 
        nxfunction (nx.algorithms.cluster.clustering): Clustering function from networkx
        colname (str): New column to add computed centrality indicators to node attribute dataframe.

    Returns:
        gpd.GeoDataFrame: Node attribute dataframe with network clustering measures included. 
    """  
    # If Graph is nx.MultiDiGraph  
    G_nodes_mod = G_nodes.copy()
    

    try:
        attr_dict = nxfunction(G, **kwargs)
    
    # else, convert to nx.MultiDiGraph
    except nx.exception.NetworkXNotImplemented:
        G2 = nx.DiGraph(G)
        attr_dict = nxfunction(G2, **kwargs)
    
    if temporal:
        T_nodes_mod = T_nodes.copy()
        T_nodes_mod[colname] = list(attr_dict.values())
        T_nodes_mod = T_nodes_mod[['osmid', colname]]
        G_nodes_mod = pd.merge(G_nodes_mod, T_nodes_mod, on='osmid', how='left')
        G_nodes_mod[colname] = G_nodes_mod[colname].fillna(0)
        G_nodes_mod[colname] = G_nodes_mod[colname].round(3)
    else:
        G_nodes_mod[colname] = list(attr_dict.values())
        G_nodes_mod[colname] = G_nodes_mod[colname].round(3)
    return G_nodes_mod


def merge_nx_property(G_nodes, nxproperty, colname, T_nodes, temporal=False, *args):
    """Add graph property from networkx graph into node attribute dataframe.

    Args:
        G_nodes (gpd.GeoDataFrame): Geopandas dataframe object with node attributes.
        nxproperty (nx.MultiDiGraph.out_degree): Out degree method of networkx graph.
        colname (str): New column to add computed centrality indicators to node attribute dataframe.

    Returns:
        gpd.GeoDataFrame: Node attribute dataframe with network property included.
    """    
    G_nodes_mod = G_nodes.copy()

    attr_dict = dict(nxproperty)
    if temporal:
        T_nodes_mod = T_nodes.copy()
        T_nodes_mod[colname] = list(attr_dict.values())
        T_nodes_mod = T_nodes_mod[['osmid', colname]]
        G_nodes_mod = pd.merge(G_nodes_mod, T_nodes_mod, on='osmid', how='left')
        G_nodes_mod[colname] = G_nodes_mod[colname].fillna(0)
    else:
        G_nodes_mod[colname] = list(attr_dict.values())
    return G_nodes_mod


import numpy as np
import pandas as pd
import pydeck as pdk
from ipywidgets import HTML
import numbers   # <- add

def get_connected_nodes(connected_nodes, connections, connection_type, node_type, node_idx):

    # Get pairs for connection type
    array_pair = connections[connection_type]

    start_node, end_node = connection_type.split('_')[0], connection_type.split('_')[-1]

    # Connections 
    if start_node == node_type:
        connected_nodes[end_node] = list(array_pair[1][np.where(array_pair[0]==node_idx)])

    if end_node == node_type:
        if node_type in connected_nodes:
            connected_nodes[start_node].extend(list(array_pair[0][np.where(array_pair[1]==node_idx)]))
        else:
            connected_nodes[start_node] = list(array_pair[0][np.where(array_pair[1]==node_idx)])

    return connected_nodes

def plot_graph(objects,
               connections,
               node_type="",
               colname="",
               node_id="",
               categorical=False):
    """
    Render a PyDeck view of an UrbanGraph and attach a colour‑bar legend.
    The legend is displayed via the `description` field of the Deck widget.
    """
    if node_type == '':
        node_type = 'plot'
        
    objects_copy = objects.copy()

    objects_copy[node_type][f'{node_type}_id'] = range(len(objects_copy[node_type]))
    
    # Colour palettes -------------------------------------------------
    colour_map_node_types = ['#177e89', '#084c61', '#db3a34', '#ffc857']
    colour_connections   = '#ff6b6b'
    colour_map_pastel    = ['#fec5bb', '#fcd5ce', '#fae1dd', '#f8edeb', '#e8e8e4',
                            '#d8e2dc', '#ece4db', '#ffe5d9', '#ffd7ba', '#fec89a']
    colour_map_vibrant   = ['#f94144', '#f3722c', '#f8961e', '#f9844a', '#f9c74f',
                            '#90be6d', '#43aa8b', '#4d908e', '#577590', '#277da1']

    # ----------------------------------------------------------------
    # Default RGBA for each layer type (used unless overridden below)
    # ----------------------------------------------------------------
    if node_type == 'plot':
        building_colour     = "[255,255,255,255]"
        plot_colour         = "[8, 76, 97,20]"
        street_colour       = "[255,255,255,255]"
        intersection_colour = "[255,255,255,255]"
    elif node_type == 'street':
        building_colour     = "[255,255,255,50]"
        plot_colour         = "[255,255,255,255]"
        street_colour       = "[23,126,137,255]"
        intersection_colour = "[255,255,255,255]"
    elif node_type == 'building':
        building_colour     = "[247,235,212,200]"
        plot_colour         = "[0,0,0,10]"
        street_colour       = "[255,255,255,255]"
        intersection_colour = "[255,255,255,255]"
    elif node_type == 'intersection':
        building_colour     = "[255,255,255,50]"
        plot_colour         = "[0,0,0,10]"
        street_colour       = "[255,255,255,255]"
        intersection_colour = "[0,0,0,255]"
    
    # Allow colour override when `colname` points at the layer ----------
    if (colname != '')        & (node_type == 'plot'):        plot_colour         = 'color'
    elif (colname != '')    & (node_type == 'street'):      street_colour       = 'color'
    elif (colname != '')  & (node_type == 'building'):    building_colour     = 'color'
    elif (colname != '') & (node_type == 'intersection'): intersection_colour = 'color'
    
    # ----------------------------------------------------------------
    # Compute fill colour column (`add_gradient_column` must exist)
    # ----------------------------------------------------------------
    if '_id' in colname:
        categorical = True  # treat *_id columns as categories

    if (colname != '') & (categorical != True):
        objects_copy[node_type][colname] = objects_copy[node_type][colname].round(1)

        objects_copy[node_type] = add_gradient_column(
            objects_copy[node_type],
            colname,
            color_stops=colour_map_pastel,
            categorical=categorical
        )
    
    # ----------------------------------------------------------------
    # Selected node & neighbours (optional)
    # ----------------------------------------------------------------
    if node_id != "":
        chosen_row = objects_copy[node_type].iloc[[node_id]].reset_index(drop=True)
        centerx, centery = (chosen_row.geometry[0].centroid.x,
                            chosen_row.geometry[0].centroid.y)
        
        chosen_layer = pdk.Layer(
            "GeoJsonLayer",
            chosen_row,
            opacity=1,
            get_fill_color='[255, 107, 107,200]',
            get_line_color='[0,0,0]',
            line_width_min_pixels=1,
        )
        
        # Collect neighbouring nodes (requires your own helper)
        relevant_edges = [k for k in connections if node_type in k and 'boundary' not in k]
        connected_nodes  = {'plot': [], 'intersection': [], 'street': [], 'building': []}
        for et in relevant_edges:
            connected_nodes = get_connected_nodes(connected_nodes,
                                                  connections,
                                                  et,
                                                  node_type,
                                                  node_id)
        neighbour_layers = []
        for ntype, neigh_idx in connected_nodes.items():
            if not neigh_idx:
                continue
            gdf = objects[ntype].iloc[neigh_idx]
            layer_args = dict(data=gdf,
                              opacity=1,
                              get_fill_color='[255, 107, 107,100]',
                              get_line_color='[0,0,0]',
                              line_width_min_pixels=1)
            if ntype == 'building':
                layer_args.update(extruded=True, get_elevation="bid_height")
            neighbour_layers.append(pdk.Layer("GeoJsonLayer", **layer_args))

    else:
        chosen_layer     = None
        neighbour_layers = []
        centerx, centery = (objects['boundary'].geometry[0].centroid.x,
                            objects['boundary'].geometry[0].centroid.y)
    
    # ----------------------------------------------------------------
    # Deck.gl layers
    # ----------------------------------------------------------------
    # pickable = True if 
    add_pickable = {"pickable": True, "auto_highlight": True}
    
    plot_layer = pdk.Layer(
        "GeoJsonLayer",
        objects_copy['plot'],
        opacity=1,
        get_fill_color=plot_colour,
        get_line_color=plot_colour,
        line_width_min_pixels=2,
        **(add_pickable if node_type == 'plot' else {})
    )
    
    street_layer = pdk.Layer(
        "GeoJsonLayer",
        objects_copy['street'],
        get_line_color=street_colour,
        line_width_min_pixels=3,
        **(add_pickable if node_type == 'street' else {})
    )
    
    intersection_layer = pdk.Layer(
        "ScatterplotLayer",
        objects_copy['intersection'],
        get_position='[x,y]',
        get_fill_color=intersection_colour,
        get_radius=4,
        **(add_pickable if node_type == 'intersection' else {})
    )
    
    building_layer = pdk.Layer(
        "GeoJsonLayer",
        objects_copy['building'],
        extruded=True,
        get_elevation="bid_height",
        get_fill_color=building_colour,
        get_line_color=building_colour,
        **(add_pickable if node_type == 'building' else {})
    )

        
    # Layer draw order
    layers = [building_layer, plot_layer, intersection_layer, street_layer]
    
    # ----------------------------------------------------------------
    # View state
    # ----------------------------------------------------------------
    view_state = pdk.ViewState(
        latitude=centery,
        longitude=centerx,
        zoom=14,
        pitch=45,
        bearing=135,
        height=1000,
        width=500,
    )

    tooltip = {
       "html": f"<b>{node_type} ID:</b> " + "{" +  f'{node_type}_id' + "}",
    }
    
    if colname != '':
        col_title = colname.replace("_", " ").title()
        tooltip = {
           "html": f"<b>{col_title}:</b> " + "{" +  f'{colname}' + "}",
        }
        
        # ----------------------------------------------------------------
        # Build legend HTML
        # ----------------------------------------------------------------
        if categorical:
            cats = (objects_copy[node_type][colname]
                    .astype(str)
                    .sort_values()
                    .unique()
                    .tolist())
            colours = colour_map_pastel[:len(cats)]
            legend_html = build_html_legend(cats, colours,
                                            title=col_title,
                                            continuous=False)
        else:
            series = objects_copy[node_type][colname]
            minmax = [series.min(), series.max()]
            legend_html = build_html_legend(minmax, colour_map_pastel,
                                            title=col_title,
                                            continuous=True)
    
    # ----------------------------------------------------------------
    # Assemble Deck
    # ----------------------------------------------------------------
    if node_id != '':
        extra_layers = [l for l in ([chosen_layer] + neighbour_layers) if l is not None]
        deck_layers  = layers + extra_layers
    else:
        deck_layers = layers
        
    deck = pdk.Deck(
        layers=deck_layers,
        initial_view_state=view_state,
        map_provider="carto",
        map_style="dark_all",
        tooltip = tooltip
    )
    
    if colname != '':
        display(deck,HTML(legend_html))    
    else:
        display(deck)


def hex_to_rgb(hex_color: str):
    """
    Convert a hex color string (e.g. '#edafb8') to an (R, G, B) tuple (0-255 each).
    """
    hex_color = hex_color.lstrip('#')
    return tuple(int(hex_color[i:i+2], 16) for i in (0, 2, 4))

def interpolate_color(color1, color2, fraction: float):
    """
    Linearly interpolate between two colors (each an (R, G, B) tuple) by the given fraction in [0,1].
    Returns an (R, G, B) tuple.
    """
    r = int(color1[0] + (color2[0] - color1[0]) * fraction)
    g = int(color1[1] + (color2[1] - color1[1]) * fraction)
    b = int(color1[2] + (color2[2] - color1[2]) * fraction)
    return (r, g, b)

def pick_color(value, min_val, max_val, color_stops):
    """
    Given a numeric value within [min_val, max_val], pick a color by piecewise linear interpolation
    among the given color stops (list of hex codes).
    """
    # Handle degenerate case (all values are the same)
    if max_val == min_val:
        return hex_to_rgb(color_stops[0])
    
    # Normalize value to a fraction t in [0, 1]
    t = (value - min_val) / (max_val - min_val)
    t = max(0, min(t, 1))  # clamp to [0, 1]
    
    num_segments = len(color_stops) - 1
    scaled_t = t * num_segments
    
    # Identify segment indices
    idx1 = int(scaled_t)
    idx2 = min(idx1 + 1, num_segments)
    
    # Fraction within this segment
    segment_fraction = scaled_t - idx1
    
    # Convert hex stops to RGB
    c1 = hex_to_rgb(color_stops[idx1])
    c2 = hex_to_rgb(color_stops[idx2])
    
    return interpolate_color(c1, c2, segment_fraction)

def add_gradient_column(
    df: pd.DataFrame,
    target_col: str,
    color_stops: list,
    new_col_name: str = 'color',
    categorical: bool = False
) -> pd.DataFrame:
    """
    Return a copy of df with a new column that contains color values (as (R,G,B) tuples).
    
    If `categorical=False`:
      - Interpolate among the given color_stops over the numeric range of `target_col`.
    
    If `categorical=True`:
      - Assign colors by repeating the color_stops in sequence for each unique category.
    """
    df_copy = df.copy()
    

    if categorical:
        # For categorical data, assign colors in a repeating sequence
        df_copy = df_copy.reset_index()
        unique_cats = df_copy[target_col].unique()
        cat_to_color = {}
        
        for i, cat in enumerate(unique_cats):
            # Cycle through color_stops by taking i % len(color_stops)
            cat_to_color[cat] = hex_to_rgb(color_stops[i % len(color_stops)])
        
        df_copy[new_col_name] = df_copy[target_col].map(cat_to_color)
    
    else:
        # For continuous data, use the numeric color interpolation
        min_val = df_copy[target_col].min()
        max_val = df_copy[target_col].max()
        
        df_copy[new_col_name] = df_copy[target_col].apply(
            lambda val: pick_color(val, min_val, max_val, color_stops)
        )
    
    return df_copy

def build_html_legend(values, colours, *, title="", continuous=False):
    """
    Return an HTML <div> containing a legend.
    See doc‑string in original code for parameters.
    """
    if continuous:                         # gradient bar ---------------------
        gradient_css = ", ".join(colours)
        min_val, max_val = values

        # helper: format only if numeric
        def _fmt(v):
            return f"{v:.2f}" if isinstance(v, numbers.Number) else str(v)

        legend = f"""
        <div style="padding:8px;background:rgba(0,0,0,0.7);border-radius:4px;
                    color:#fff;font-family:Arial;font-size:11px;">
          <div style="font-weight:bold;margin-bottom:4px;">{title}</div>
          <div style="display:flex;align-items:center;gap:4px;">
            <span>{_fmt(min_val)}</span>
            <div style="flex:1;height:12px;background:
                        linear-gradient(to right,{gradient_css});"></div>
            <span>{_fmt(max_val)}</span>
          </div>
        </div>"""
    else:                                  # discrete swatches ---------------
        rows = "".join(
            f"<div style='display:flex;align-items:center;margin-bottom:2px;'>"
            f"  <div style='width:12px;height:12px;background:{c};"
            f"       margin-right:4px;'></div>{v}</div>"
            for v, c in zip(values, colours)
        )
        legend = f"""
        <div style="padding:8px;background:rgba(0,0,0,0.7);border-radius:4px;
                    color:#fff;font-family:Arial;font-size:11px;">
          <div style="font-weight:bold;margin-bottom:4px;">{title}</div>
          {rows}
        </div>"""
    return legend.strip()



import os
from dataclasses import dataclass, field
import geopandas as gpd
import pandas as pd
import numpy as np
import pickle
import torch
import requests
from .utils import get_plot_to_plot_edges, get_building_to_street_edges, get_edge_nodes, get_building_to_building_edges, \
                   get_intersection_to_street_edges, get_buildings_in_plot_edges, get_edges_along_plot, boundary_to_plot, \
                   remove_non_numeric_columns_objects, standardise_and_scale, fill_na_in_objects, one_hot_encode_categorical
from typing import Dict
from .visualisation import plot_graph
from torch_geometric.data import HeteroData
import torch_geometric.transforms as T

import zipfile
import io

@dataclass
class UrbanGraph:
    boundary: gpd.GeoDataFrame = field(default_factory=lambda: gpd.GeoDataFrame())
    building: gpd.GeoDataFrame = field(default_factory=lambda: gpd.GeoDataFrame())
    plot: gpd.GeoDataFrame = field(default_factory=lambda: gpd.GeoDataFrame())
    street: gpd.GeoDataFrame = field(default_factory=lambda: gpd.GeoDataFrame())
    intersection: gpd.GeoDataFrame = field(default_factory=lambda: gpd.GeoDataFrame())

    boundary_to_plot: np.ndarray = field(default_factory=lambda: np.array('None'))
    plot_to_boundary: np.ndarray = None
    plot_to_plot: np.ndarray = None
    building_to_building: np.ndarray = None
    building_to_street: np.ndarray = None
    street_to_building: np.ndarray = None
    intersection_to_street: np.ndarray = None
    street_to_intersection: np.ndarray = None
    building_to_plot: np.ndarray = None
    plot_to_building: np.ndarray = None
    street_to_plot: np.ndarray = None
    plot_to_street: np.ndarray = None

    # We'll defer populating these until after the instance is created:
    geo_store: Dict = field(default_factory=dict, init=False)
    edge_store: Dict = field(default_factory=dict, init=False)

    def __post_init__(self):
        """Called automatically after dataclass __init__ to populate geo_store and edge_store."""
        self.geo_store = {
            'boundary': self.boundary,
            'building': self.building,
            'plot': self.plot,
            'street': self.street,
            'intersection': self.intersection
        }

        self.edge_store = {
            'boundary_to_plot': self.boundary_to_plot,
            'plot_to_boundary': self.plot_to_boundary,
            'plot_to_plot': self.plot_to_plot,
            'building_to_building': self.building_to_building,
            'building_to_street': self.building_to_street,
            'street_to_building': self.street_to_building,
            'intersection_to_street': self.intersection_to_street,
            'street_to_intersection': self.street_to_intersection,
            'building_to_plot': self.building_to_plot,
            'plot_to_building': self.plot_to_building,
            'street_to_plot': self.street_to_plot,
            'plot_to_street': self.plot_to_street
        }

    def __repr__(self) -> str:
        """
        Return a string representation of the UrbanGraph, summarizing each property.
        """

        # Build strings for each
        info1 = [self.size_repr(k, v, 2) for k, v in self.geo_store.items()]
        info2 = [self.size_repr(k, v, 2) for k, v in self.edge_store.items()]

        info = ',\n'.join(info1 + info2)
        # Optionally add line breaks if not empty
        info = f'\n{info}\n' if info else ''

        return f'{self.__class__.__name__}({info})'
    
    def size_repr(self, name, data, indent=2):
        """
        Return a string that describes the size or shape of data.
        Indent controls how many spaces prefix the line.
        """
        prefix = ' ' * indent
        if data is None:
            return f"{prefix}{name}: None"
        if isinstance(data, gpd.GeoDataFrame):
            return f"{prefix}{name}: GeoDataFrame with {len(data)} rows"
        elif isinstance(data, pd.DataFrame):
            return f"{prefix}{name}: DataFrame with {len(data)} rows, {data.shape[1]} columns"
        elif isinstance(data, np.ndarray):
            return f"{prefix}{name}: ndarray shape {data.shape}"
        else:
            # Fallback for other types
            return f"{prefix}{name}: {type(data).__name__}"
        
    def initialize_edges(self, building_neighbours = 'knn', knn=5, distance=100):

        if self.boundary_to_plot.size == 1: 
            _, self.plot_to_plot = get_plot_to_plot_edges(self.plot)
            _, self.building_to_building = get_building_to_building_edges(self.building, 
                                                                        return_neighbours = building_neighbours,
                                                                        knn = knn,
                                                                        distance_threshold = distance,
                                                                        knn_threshold = distance, 
                                                                        add_reverse=True)
            
            self.boundary_to_plot, self.plot_to_boundary = boundary_to_plot(self.plot)
            self.street_to_plot, self.plot_to_street = get_edges_along_plot(self.plot)
            self.building_to_street, self.street_to_building = get_building_to_street_edges(self.street, self.building)
            self.intersection_to_street, self.street_to_intersection = get_intersection_to_street_edges(self.intersection, self.street)
            self.building_to_plot, self.plot_to_building = get_buildings_in_plot_edges(self.plot)
            self.street_to_plot, self.plot_to_street = get_edges_along_plot(self.plot)

            self.geo_store = {
                'boundary': self.boundary,
                'building': self.building,
                'plot': self.plot,
                'street': self.street,
                'intersection': self.intersection
            }

            self.edge_store = {
                'boundary_to_plot': self.boundary_to_plot,
                'plot_to_boundary': self.plot_to_boundary,
                'plot_to_plot': self.plot_to_plot,
                'building_to_building': self.building_to_building,
                'building_to_street': self.building_to_street,
                'street_to_building': self.street_to_building,
                'intersection_to_street': self.intersection_to_street,
                'street_to_intersection': self.street_to_intersection,
                'building_to_plot': self.building_to_plot,
                'plot_to_building': self.plot_to_building,
                'street_to_plot': self.street_to_plot,
                'plot_to_street': self.plot_to_street
            }

            # Process for downstream tasks
            self.geo_store = remove_non_numeric_columns_objects(self.geo_store, keep_geometry=True)
        else:
            print('Graph edges already initialized.')

    # Any additional methods (including __repr__) go here
    def plot_urban_graph(self, node_type="", colname="", node_id=""):
        graph_viz = plot_graph(self.geo_store, 
                               self.edge_store, 
                               node_type=node_type,
                               colname=colname,
                               node_id=node_id,
                               categorical=False)
        
    def save_graph(self, filename):
        with zipfile.ZipFile(filename, 'w') as z:
            # Save each GeoDataFrame as a GeoParquet file inside the zip
            for name, gdf in self.geo_store.items():
                if name == 'plot':
                    target_cols = ['street_id', 'bid']
                    for col in target_cols:
                        npy_name = f"plot_{col}.npy"
                        arr = np.array(self.geo_store[name][col].to_list(), dtype=object)
                        np.save(npy_name, arr, allow_pickle=True)
                        z.write(npy_name, arcname=npy_name)
                        os.remove(npy_name)  # clean up
                        gdf = gdf.drop(columns=col)
                parquet_name = f"{name}.parquet"
                gdf.to_parquet(parquet_name)
                z.write(parquet_name, arcname=parquet_name)
                os.remove(parquet_name)  # clean up

            # Save each numpy array
            for name, arr in self.edge_store.items():
                npy_name = f"{name}.npy"
                np.save(npy_name, arr)
                z.write(npy_name, arcname=npy_name)
                os.remove(npy_name)  # clean up

    def load_graph(self, filename):
        """
        Reads each .parquet and .npy file in the zip, storing them in two dictionaries.
        Supports both local and online ZIP files.
        Returns:
            geodf_dict: dict of {filename_without_ext: GeoDataFrame}
            array_dict: dict of {filename_without_ext: numpy.ndarray}
        """
        geodf_dict = {}
        array_dict = {}

        # If filename is a URL, fetch into memory
        if isinstance(filename, str) and filename.startswith(('http://', 'https://')):
            response = requests.get(filename)
            response.raise_for_status()
            zip_data = io.BytesIO(response.content)
            zip_file = zipfile.ZipFile(zip_data, 'r')
        else:
            zip_file = zipfile.ZipFile(filename, 'r')

        with zip_file as z:
            # List all files in the ZIP archive
            for file in z.namelist():
                # If it's a GeoDataFrame in Parquet format
                if file.endswith('.parquet'):
                    with z.open(file) as f:
                        gdf = gpd.read_parquet(f)
                    dict_key = file.rsplit('.', 1)[0]
                    geodf_dict[dict_key] = gdf

            for file in z.namelist():
                # If it's a NumPy array
                if file.endswith('.npy'):
                    if 'id' in file:
                        with z.open(file) as f:
                            data = io.BytesIO(f.read())
                            arr = np.load(data, allow_pickle=True)
                            dict_key = file.rsplit('.', 1)[0][5:]
                            geodf_dict['plot'][dict_key] = arr
                    else:
                        with z.open(file) as f:
                            data = io.BytesIO(f.read())
                            arr = np.load(data, allow_pickle=True)
                        dict_key = file.rsplit('.', 1)[0]
                        array_dict[dict_key] = arr

        # Assign attributes as before
        self.boundary = geodf_dict.get('boundary', None)
        self.building = geodf_dict.get('building', None)
        self.plot = geodf_dict.get('plot', None)
        self.street = geodf_dict.get('street', None)
        self.intersection = geodf_dict.get('intersection', None)

        self.boundary_to_plot = array_dict.get('boundary_to_plot', None)
        self.plot_to_boundary = array_dict.get('plot_to_boundary', None)
        self.plot_to_plot = array_dict.get('plot_to_plot', None)
        self.building_to_building = array_dict.get('building_to_building', None)
        self.building_to_street = array_dict.get('building_to_street', None)
        self.street_to_building = array_dict.get('street_to_building', None)
        self.intersection_to_street = array_dict.get('intersection_to_street', None)
        self.street_to_intersection = array_dict.get('street_to_intersection', None)
        self.building_to_plot = array_dict.get('building_to_plot', None)
        self.plot_to_building = array_dict.get('plot_to_building', None)
        self.street_to_plot = array_dict.get('street_to_plot', None)
        self.plot_to_street = array_dict.get('plot_to_street', None)

        self.geo_store = geodf_dict if geodf_dict else None
        self.edge_store = array_dict if array_dict else None

    def generate_masks(self, df_length: int,
                       train_ratio: float = 0.6,
                       val_ratio: float = 0.2,
                       test_ratio: float = 0.2,
                       random_state: int = 42):
        """
        Generate boolean masks for train, validation, and test splits given 
        the number of rows in a GeoDataFrame and the desired split ratios.
        """
        if abs(train_ratio + val_ratio + test_ratio - 1.0) > 1e-6:
            raise ValueError("Train, validation, and test ratios must sum to 1.")

        indices = np.arange(df_length)
        rng = np.random.default_rng(random_state)
        rng.shuffle(indices)

        train_end = int(train_ratio * df_length)
        val_end = train_end + int(val_ratio * df_length)

        train_idx = indices[:train_end]
        val_idx = indices[train_end:val_end]
        test_idx = indices[val_end:]

        train_mask = np.zeros(df_length, dtype=bool)
        val_mask = np.zeros(df_length, dtype=bool)
        test_mask = np.zeros(df_length, dtype=bool)

        train_mask[train_idx] = True
        val_mask[val_idx] = True
        test_mask[test_idx] = True

        return train_mask, val_mask, test_mask

    def generate_building_masks(self,
                                train_ratio: float = 0.6,
                                val_ratio: float = 0.2,
                                test_ratio: float = 0.2,
                                random_state: int = 42):
        """
        Generate train, validation, and test masks for the building GeoDataFrame.
        """
        return self.generate_masks(len(self.building),
                                   train_ratio,
                                   val_ratio,
                                   test_ratio,
                                   random_state)

    def generate_plot_masks(self,
                            train_ratio: float = 0.6,
                            val_ratio: float = 0.2,
                            test_ratio: float = 0.2,
                            random_state: int = 42):
        """
        Generate train, validation, and test masks for the plot GeoDataFrame.
        """
        return self.generate_masks(len(self.plot),
                                   train_ratio,
                                   val_ratio,
                                   test_ratio,
                                   random_state)

    def generate_street_masks(self,
                              train_ratio: float = 0.6,
                              val_ratio: float = 0.2,
                              test_ratio: float = 0.2,
                              random_state: int = 42):
        """
        Generate train, validation, and test masks for the street GeoDataFrame.
        """
        return self.generate_masks(len(self.street),
                                   train_ratio,
                                   val_ratio,
                                   test_ratio,
                                   random_state)

    def generate_intersection_masks(self,
                                    train_ratio: float = 0.6,
                                    val_ratio: float = 0.2,
                                    test_ratio: float = 0.2,
                                    random_state: int = 42):
        """
        Generate train, validation, and test masks for the intersection GeoDataFrame.
        """
        return self.generate_masks(len(self.intersection),
                                   train_ratio,
                                   val_ratio,
                                   test_ratio,
                                   random_state)

    def to_pyg_graph(self, 
                     target_node = 'building', 
                     categorical = False,
                     target_value = [], 
                     train_val_test=[0.6, 0.2, 0.2],
                     random_seed=0
                     ):

        data = HeteroData()
        objects_copy = self.geo_store.copy()

        if target_value:
            assert len(target_value) == len(objects_copy[target_node]), "Groundtruth labels do not match length of entity. Missing values should be specified as None."

        node_types = ['boundary', 'plot','building','street','intersection']

        for node in node_types:
            objects_copy[node][f'{node}_id'] = range(len(objects_copy[node]))
            objects_copy[node] = objects_copy[node].drop(columns = ['geometry'], axis=1)

        from sklearn.preprocessing import StandardScaler
        from sklearn.compose import ColumnTransformer

        scale = StandardScaler()

        # Standardize data
        cols = [i for i in objects_copy['plot'].columns if not i.startswith('lcz_')]
        ct = ColumnTransformer([
                ('somename', StandardScaler(), cols)
            ], remainder='passthrough')
        
        # Rescale all nodes
        objects_copy['boundary'] = scale.fit_transform(objects_copy['boundary'])
        objects_copy['street'] = scale.fit_transform(objects_copy['street'])
        objects_copy['intersection'] = scale.fit_transform(objects_copy['intersection'])
        objects_copy['building'] = scale.fit_transform(objects_copy['building'])
        objects_copy['plot'] = ct.fit_transform(objects_copy['plot'])

        node_types = ['boundary', 'plot','building','street','intersection']

        for node in node_types:
            data[node].x = torch.from_numpy(objects_copy[node].astype(np.float32))
                
        # Insert edges
        for key, arr in self.edge_store.items():
            splitted = key.split('_')
            if 'rev' in key:
                data[splitted[0], 'rev_to', splitted[-1]].edge_index = torch.from_numpy(arr.copy()).to(torch.int64)
            else:
                data[splitted[0], 'to', splitted[2]].edge_index = torch.from_numpy(arr.copy()).to(torch.int64)

        if target_value:
            if categorical:
                arr = np.array(target_value)
                # 1. Find unique categories (excluding None)
                categories = sorted(set([x for x in arr if x is not None]))
                cat2int = {cat: idx for idx, cat in enumerate(categories)}

                print(f'Using categorical mapping: {cat2int}')
                # 2. Map array values, assign -1 to None
                def encode(val):
                    return cat2int.get(val, -1)

                int_arr = np.array([encode(x) for x in arr])
                data[target_node].y = torch.from_numpy(np.array(int_arr))
            else:
                arr = np.array(target_value)
                arr = np.nan_to_num(arr, nan=-1)
                data[target_node].y = torch.from_numpy(arr)
            
        else:
            data[target_node].y = torch.from_numpy(np.array(np.random.randint(0,5,len(objects_copy[target_node]))))

        import random
        # Set train, validation, and test masks
        torch.manual_seed(random_seed)
        random.seed(random_seed)
        np.random.seed(random_seed)

        valid_indices = np.where(data[target_node].y != -1)[0]
        np.random.shuffle(valid_indices)

        # Set your desired split ratios
        train_ratio, val_ratio, test_ratio = train_val_test

        # Compute split sizes
        train_end = int(train_ratio * len(valid_indices))
        val_end = train_end + int(val_ratio * len(valid_indices))

        # Create index splits
        train_idx = valid_indices[:train_end]
        val_idx = valid_indices[train_end:val_end]
        test_idx = valid_indices[val_end:]

        # Instantiate empty vector masks
        train_mask = np.zeros(len(objects_copy[target_node])).astype(int)
        val_mask = np.zeros(len(objects_copy[target_node])).astype(int)
        test_mask = np.zeros(len(objects_copy[target_node])).astype(int)

        # Assign 1 to entries corresponding to each index
        np.put(train_mask, train_idx, 1)
        np.put(val_mask, val_idx, 1)
        np.put(test_mask, test_idx, 1)

        # Insert train, val, and test masks
        data[target_node].train_mask = torch.from_numpy(train_mask).bool()
        data[target_node].val_mask = torch.from_numpy(val_mask).bool()
        data[target_node].test_mask = torch.from_numpy(test_mask).bool()

        data = T.AddSelfLoops()(data)

        return data

    def to_dgl():
        pass
    